#!/usr/bin/env python

import sys
import math, numpy, os
import lsst.daf.persistence as dafPersist
import lsst.obs.hsc as obsHsc
import lsst.afw.image as afwImage
import lsst.afw.geom as afwGeom

import hsc.qa.hscTrail as hscTrail

def main():

    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument('inputFile', help = 'input CORR FITS file to mask trails')
    parser.add_argument('outputFile', help = 'prefix of output FITS file name')
    parser.add_argument('--do-write-file', dest = 'doWriteFile', action = 'store_true', help = 'writing debug files')
    parser.add_argument('--min-hough', dest = 'minHough', type = int, default = 1000, help='minimum peak count in Hough phase space to recognize as a peak. if --do-scale-minhoug is set, this option is ignored (default: 1000)')
    parser.add_argument('--min-peak-area', dest = 'minPeakArea', type = int, default = 2, help='minimum connected pixels above the peak count in Hough phase space to recognize as a peak (default: 2)')
    parser.add_argument('--min-peak-counts', dest = 'minPeakCounts', type = int, default = -1, help='minimum sum of peak counts over the pixels within peak-area-radius1 above the min-hough count in Hough phase space to recognize as a peak (default: 10000)')
    parser.add_argument('--peak-area-radius1', dest = 'rMargin1', type = float, default = 10.0, help='inner margin radius to count up peak pixels in hough phase space (default: 10.0)')
    parser.add_argument('--peak-area-radius2', dest = 'rMargin2', type = float, default = 60.0, help='outer margin radius for grouping individual peak pixels in hough phase space. pixels within this radius belongs to the same peak and will not go to another neighboring peak  (default: 60.0)')
    parser.add_argument('--allowed-peak-ratio', dest = 'allowedPeakRatio', type = float, default = -1.0, help='all the Hough peaks should be larger than the first peak devided by this ratio (default: unset)')
    parser.add_argument('--min-trail-area', dest = 'minTrailArea', type = int, default = 300, help='minimum area of each trail candiate (default: 300)')
    parser.add_argument('--trail-margin', dest = 'trailMargin', type = float, default = 30.0, help='acceptable spatial separation of actual trails to be masksed from the estimated line features (default: 30.0)')
    parser.add_argument('--bin', dest = 'binSize', type = int, default = 1, help='image binning factor for speed up processing time (default: 1)')
    parser.add_argument('--lsigma', dest = 'lowSigma', type = float, default = 2.0, help='nsigma for image low-threshold for Hough transformation (default: 2.0)')
    parser.add_argument('--hsigma', dest = 'highSigma', type = float, default = -1.0, help='nsigma for image high-threshold for Hough transformation (default: None)')
    parser.add_argument('--min-image', dest = 'minImage', type = float, default = None, help='minimum count for image low-threshold for Hough transformation. This option overrides --lsigma (default: unset)')
    parser.add_argument('--max-image', dest = 'maxImage', type = float, default = None, help='maximum count for image high-threshold for Hough transformation. This option overrides --hsigma (default: unset)')
    parser.add_argument('--min-hough-footprint', dest = 'minHoughPerFootprint', type = int, default = -1, help = 'min-hough per footprint. if this is set --min-hough will be ignored.')
    parser.add_argument('--skip-narrow-footprints', dest = 'doNarrowFootprints', action = 'store_false', help='do squash footprints before hough transformation.')
    parser.add_argument('--line-width', dest = 'lineWidth', type = float, default = 10.0, help='width to narrow footprints before Hough transformation. (default: 10.0)')
    parser.add_argument('--keep-round-footprints', dest = 'doDropRoundFootprints', action = 'store_false', help='drop relatively round footprints before hough transformation. (default: false)')
    parser.add_argument('--max-det2bbox', dest = 'maxDetToBBoxRatio', type = float, default = 0.35, help='maximum ratio of detected area and footprint bbox area not to be excluded in drop round footprint.')
    parser.add_argument('--max-aratio', dest = 'maxFootprintAspectRatio', type = float, default = 5.0, help = 'maximum aspect ratio (x/y) of each footprint to be narrowed in both x and y directions.')
    parser.add_argument('--min-aratio', dest = 'minFootprintAspectRatio', type = float, default = 1.5, help = 'minimum aspect ratio (x/y) of a footprint not to be excluded in drop round footprint.')
    args = parser.parse_args()
    inputFile = args.inputFile
    outputFile = args.outputFile

#    file = 'CORR-0022374-034.fits'
#    file = 'CORR-0013200-090.fits'
#    file = 'CORR-0013200-091.fits'
#    file = 'CORR-0013200-096.fits'
#    file = 'CORR-0013200-097.fits' # many lines are detected
#    file = 'CORR-0013202-098.fits'
#    file = 'CORR-0013202-099.fits'

#    nAngle = 16
#    nRho = 32
###    nAngle = 200
###    nRho = 200


    exp = afwImage.ExposureF(inputFile)

    trail = hscTrail.HscTrailTask(hscTrail.HscTrailConfig())

    config = trail.config

    config.minHough = args.minHough
    config.minPeakArea = args.minPeakArea
    config.minPeakCounts = args.minPeakCounts
    config.allowedPeakRatio = args.allowedPeakRatio
    config.minTrailArea = args.minTrailArea
    config.trailMargin = args.trailMargin
    config.binSize = args.binSize
    config.rMargin1 = args.rMargin1
    config.rMargin2 = args.rMargin2

    config.maxDetToBBoxRatio = args.maxDetToBBoxRatio
    config.maxFootprintAspectRatio = args.maxFootprintAspectRatio
    config.minFootprintAspectRatio = args.minFootprintAspectRatio
    config.minHoughPerFootprint = args.minHoughPerFootprint
    config.narrowedLineWidth = args.lineWidth
    config.doDropRoundFootprints = args.doDropRoundFootprints
    config.doNarrowFootprints = args.doNarrowFootprints
    config.doWriteFile = args.doWriteFile

    config.outputFile = args.outputFile

    #
    # detect all possible trails
    #
    trail.run(exp, display = True, debug = True)
#    trail.run(exp, display = False, debug = False)

    exp.writeFits(outputFile + '_masked.fits')


if __name__ == '__main__':
    main()


