#!/usr/bin/env python 
import numpy, sys 

import lsst.afw.image as afwImage
import lsst.afw.detection as afwDetection
import lsst.afw.display.ds9 as ds9
import lsst.afw.math as afwMath
import hsc.qa as qa


class TrailsTask(object):

    def __init__(self, **kw):

        self.nAngle = 500
        self.nRho = 500
        self.rMargin = 60

        self.minHough = 7000
        self.minPeakArea = 10
        self.allowedPeakRatio = 1.5
        self.minTrailArea = 300
        self.trailMargin =30.0

        self.binSize = 1
        
        if kw['nAngle']: self.nAngle = kw['nAngle']
        if kw['nRho']: self.nRho = kw['nRho']
        if kw['rMargin']: self.rMargin = kw['rMargin']
        if kw['minHough']: self.minHough = kw['minHough']
        if kw['minPeakArea']: self.minPeakArea = kw['minPeakArea']
        if kw['allowedPeakRatio']: self.allowedPeakRatio = kw['allowedPeakRatio']
        if kw['minTrailArea']: self.minTrailArea = kw['minTrailArea']
        if kw['trailMargin']: self.trailMargin = kw['trailMargin']
        if kw['binSize']: self.binSize = kw['binSize']

    def getTrailCandidates(self, exp, lowSigma = None, highSigma = None, minImage = None, maxImage = None):
        """
        get possible trail candidates based on Hough transformation
        """

        mi = exp.getMaskedImage()
        mask = mi.getMask()

        # binning mode is not extensively tested
        if self.binSize > 1:
            print 'doing binning by: %d x %d' % (binSize, binSize)
            miBinned = afwMath.binImage(mi, self.binSize)
            # sampling due to rho space can be reduced thanks to binning, but angle space keeps the same
            himage = self.getHoughImage(miBinned, goodMask = None, badMask = None, 
                                        nRho = int(float(self.nRho) / self.binSize), 
                                        nAngle = self.nAngle, 
                                        lowSigma = lowSigma, highSigma = highSigma, 
                                        minImage = minImage, maxImage = maxImage
                                        )
            himage = afwMath.binImage(himage, self.binSize, 1, afwMath.MEAN) # binning in angle space to use rMargin in the same manner as in the unbinned case
            xHoughPeakList, yHoughPeakList = self.getHoughPeaks(maskBinned, himage, 
                                                                minHough = (self.minHough / self.binSize / self.binSize), 
                                                                allowedPeakRatio = self.allowedPeakRatio, 
                                                                minPeakArea = (self.minPeakArea / self.binSize / self.binSize),
                                                                rMargin = (self.rMargin / self.binSize)
                                                                )
            # convert xPeak, yPeak into what would be on the original unbinned himage
            xHoughPeakList, yHoughPeakList = xHoughPeakList * float(self.binSize), yHoughPeakList * float(self.binSize)
        else:
            himage = self.getHoughImage(mi, goodMask = None, badMask = None, 
                                        nRho = self.nRho, nAngle = self.nAngle,
                                        lowSigma = lowSigma, highSigma = highSigma, 
                                        minImage = minImage, maxImage = maxImage
                                        )
            xHoughPeakList, yHoughPeakList = self.getHoughPeaks(mask, himage, 
                                                                minHough = self.minHough, 
                                                                allowedPeakRatio = self.allowedPeakRatio, 
                                                                minPeakArea = self.minPeakArea, 
                                                                rMargin = self.rMargin)        

        aList, bList = self.convertHoughToTrails(mask, xHoughPeakList, yHoughPeakList, 
                                                 nRho = self.nRho, nAngle = self.nAngle
                                                 )

        if False:
            himage.writeFits('hough_output.fits')

        # display the resultant trails on ds9
        if True:
            ds9.mtv(exp.getMaskedImage().getImage(), frame = 1) 
        for (a, b) in zip(aList, bList):
            xx = numpy.array([0, mask.getWidth()-1]) 
            yy = a * xx + b 
            points = numpy.array([(x, y) for (x, y) in zip(xx, yy)])
            ds9.line(points, frame = 1, ctype='cyan')

        return aList, bList

        
    def getHoughImage(self, mi, goodMask = None, badMask = None, nRho = 500, nAngle = 500, 
                      lowSigma = None, highSigma = None, minImage = None, maxImage = None
                      ):
        """
        perform Hough transformation and get Hough image
        """

        mask = mi.getMask()

        if goodMask is None:
            goodMask = mask.getPlaneBitMask("DETECTED")
        if badMask is None:
            badMaskList = ['NO_DATA', 'EDGE', 'BAD', 'SAT']
            #badMaskList = ['NO_DATA', 'EDGE', 'BAD']
            #badMaskList = ['NO_DATA', 'EDGE', 'BAD', 'DETECTED']
            #badMaskList = ["NO_DATA", "EDGE", "SAT", "SUSPECT", "INTRP", "BAD"]
            badMask = mask.getPlaneBitMask(badMaskList)

        print 'goodMaskBit = %d, badMaskBit = %d' % (goodMask, badMask)


        # calculate skysigma and threshold for pixels used in Hough
        if minImage is None or maxImage is None:
            maskVal = 0
            for maskBit in badMaskList:
                maskVal |= afwImage.MaskU.getPlaneBitMask(maskBit)
            nClip = 3.0; nIter = 3
            sctrl = afwMath.StatisticsControl(nClip, nIter, maskVal)
            stat = afwMath.makeStatistics(mi, afwMath.MEANCLIP | afwMath.STDEVCLIP | afwMath.MAX, sctrl)
            bkg = stat.getValue(afwMath.MEANCLIP)
            bkgSigma = stat.getValue(afwMath.STDEVCLIP)

            if minImage is None:
                if lowSigma is not None and lowSigma > 0:
                    minImage = bkg + lowSigma * bkgSigma
            if maxImage is None:
                if highSigma is not None and highSigma > 0:
                    maxImage = bkg + highSigma * bkgSigma
                else:
                    maxImage = stat.getValue(afwMath.MAX)

        else: # use the given minImage and maxImage
            pass
    
        #
        # perform Hough transformation
        #
        maskArr = mask.getArray()
        maskArr = numpy.array(maskArr, dtype=numpy.int32)
        maskHough = afwImage.ImageI(maskArr)

        print 'nAngle:', nAngle, 'nRho:', nRho
        hough = qa.Hough(maskHough, nAngle, nRho)

        if minImage is not None:
            print 'image count threshold applied to Hough transform: (%s, %s) (count)' % (minImage, maxImage)
            im = mi.getImage()
            himage, ang, dis = hough.calcHough(im, minImage, maxImage, goodMask, badMask)
        else:
            print 'image count threshold not applied to Hough transform'
            himage, ang, dis = hough.calcHough(goodMask, badMask)

        print 'Npix effective: ', hough.getNpixEffective()

        self.himage = himage

        # This is the highest peak returned by in Hough class 
        # But we need to detect all possible peaks corresponding to trails, so do not simply use this
        a = -1.0 * (numpy.cos(ang)/numpy.sin(ang))
        b = dis / numpy.sin(ang)
        ####    print ang, dis
        ####    print "y = %g * x + %g" % (a, b)

        return self.himage


    def getHoughPeaks(self, mask, himage, minHough, allowedPeakRatio = 1.5, 
                      minPeakArea = 10, rMargin = 60.0
                      ):
        """
        select all possible peaks (x, y) in resultant Hough-transformed plane
        """

        # this is a piece of experiments of scaling conditions 
        # based on 2k4k CORR experience to other-size data
        if False:
            maskArr = mask.getArray()
            maskEff = maskArr[ 
                numpy.logical_and((maskArr & mask.getPlaneBitMask("NO_DATA") == 0), 
                                  (maskArr & mask.getPlaneBitMask("EDGE") == 0)) 
                ]

            scale = float(len(maskEff)) / (2048.*4176.)
            if scale < 1: # if the effective area is smaller than 2k4k, we will use 1 as the lowest scale
                scale = 1.0
            minHough *= scale # scaled by effective image area
            print 'effective area: %d (pix) / (2048*4176) = %8.4f : scale_used = %8.4f' % (len(maskEff), float(len(maskEff)) / (2048.*4176.), scale)

        vHough = himage.getArray()
        nyHough, nxHough = vHough.shape
        print 'Hough nx: %d ny: %d' % (nxHough, nyHough)

        # convert hough to 1d array 
        vHough1d = numpy.reshape(vHough, nxHough * nyHough )
        x1dInd = numpy.array( [range(nxHough)] * nyHough, dtype=numpy.int32 )
        y1dInd = numpy.array( [ [y] * nxHough for y in range(nyHough) ], dtype=numpy.int32 )
        x1dInd = numpy.reshape(x1dInd, nxHough * nyHough )
        y1dInd = numpy.reshape(y1dInd, nxHough * nyHough )

        # pick out only high hough pixels
        nCand = len(vHough1d)
        print 'candidates before filtering:', nCand
        if False:
            indHigh = numpy.where(vHough1d > minHough)
        else: # workaround
            indHigh = numpy.logical_and(vHough1d > minHough, y1dInd != 500)
            vHough1dHigh = vHough1d[indHigh]
            x1dIndHigh = x1dInd[indHigh]
            y1dIndHigh = y1dInd[indHigh]

        # grouping high hough-pixel knots, 
        #   assigning one pixel for each knot group by taking the highest hough value in a knot
        xCandList = numpy.array([])
        yCandList = numpy.array([])

        if len(vHough1dHigh) == 0:
            print 'No trail candidate found'
            # display the resultant Hough peaks on ds9
            if True:
                ds9.mtv(himage, frame = 0)
            return xCandList, yCandList 

        ind1dHighSorted = numpy.argsort(vHough1dHigh)
        #   revert to make it descending
        ind1dHighSorted = ind1dHighSorted[-1::-1]
        vHough1dHighSorted = vHough1dHigh[ind1dHighSorted]
        x1dIndHighSorted = x1dIndHigh[ind1dHighSorted]
        y1dIndHighSorted = y1dIndHigh[ind1dHighSorted]

        # further limit pixels by the hough's peak ratio
        maxPeak = vHough1dHighSorted[0]
        minPeak = maxPeak / allowedPeakRatio
        print 'allowedPeakRatio = %f' %  allowedPeakRatio
        print 'maxPeak = %f' % maxPeak
        print 'minPeak = %f' % minPeak
        if False:
        #if allowedPeakRatio > 1:
            indLim = numpy.where(vHough1dHighSorted > (peakMax/allowedPeakRatio))
            vHough1dFiltered = vHough1dHighSorted[indLim]
            x1dIndFiltered = x1dIndHighSorted[indLim]
            y1dIndFiltered = y1dIndHighSorted[indLim]
        else:
            vHough1dFiltered = vHough1dHighSorted
            x1dIndFiltered = x1dIndHighSorted
            y1dIndFiltered = y1dIndHighSorted

        #   start grouping knots, with descending-sorted vHoughArr
        vHoughArr = numpy.copy(vHough1dFiltered)
        xArr = numpy.copy(x1dIndFiltered)
        yArr = numpy.copy(y1dIndFiltered)

        while len(vHoughArr) > 0:
            peakThisKnot = vHoughArr[0] # highest peak in this group
            x0 = xArr[0]
            y0 = yArr[0]
            
            # count up hough points which belongs to this knot group, and
            # carry over the other groups to the next loop
            nPixelsThisKnot = 0
            indOthers = numpy.array([], dtype = int)
            for i, (x, y) in enumerate(zip(xArr, yArr)):
                if numpy.hypot(x - x0, y - y0) < rMargin:
                    nPixelsThisKnot += 1
                else:
                    indOthers = numpy.append(indOthers, i)

            # decide if this group is a significant trail
            if (peakThisKnot > minPeak) and (nPixelsThisKnot >= minPeakArea):
                xCandList = numpy.append(xCandList, x0)
                yCandList = numpy.append(yCandList, y0)
                print 'thisKnot(peak, npixel): (%s, %s)' % (peakThisKnot, nPixelsThisKnot)

            vHoughArr = vHoughArr[indOthers]
            xArr = xArr[indOthers]
            yArr = yArr[indOthers]

        nCandHigh = len(xCandList)
        print 'candidates after filtering:', nCandHigh


        # display the resultant Hough peaks on ds9
        if True:
            ds9.mtv(himage, frame = 0)
            for (t, r) in zip(xCandList, yCandList):
                ds9.dot('o', t, r, frame = 0, size = rMargin, ctype = 'cyan')


        return xCandList, yCandList


    def convertHoughToTrails(self, mask, xHoughPeakList, yHoughPeakList, nRho, nAngle):
        """
        convert input Hough peaks into line parameters (a:slope, b:intersect)
        """
    
        nxImage, nyImage = mask.getDimensions()

        anglePeakList = numpy.pi * xHoughPeakList / nAngle
        rhoPeakList = (yHoughPeakList - float(nRho)) * int(numpy.hypot(nxImage-1, nyImage-1) + 1) / float(nRho)
        
        aList = -1.0 * (numpy.cos(anglePeakList)/numpy.sin(anglePeakList))
        bList = rhoPeakList / numpy.sin(anglePeakList)

        # !!!
        # better to handle cases of angle = 0 or a = inf

        return aList, bList


    def getTrailFootprints(self, mask, fpList, aList, bList, trailMargin, minTrailArea):
        """ select footprints of trail candidates from the input footprint list
        """

# check if this footprint overlaps with the trails candidates in each y 

        xmin, ymin = mask.getBBox(afwImage.PARENT).getMin()

        fpTrailList = []
        for fp in fpList:

            # footprint should be larger than the given minArea to be a trail
            if False:
                print '*** minTrailArea:', fp.getArea()
            if fp.getArea() < minTrailArea:
                continue

            # check if this fp overlaps with trail candidates
            for s in fp.getSpans():
                x0, x1 = s.getX0() -xmin, s.getX1() -xmin
                y = s.getY() - ymin

                # define trail regions with margin for recognizing this footprint as a candidate
                xTrailList = (y - bList) / aList
                minXTrailList, maxXTrailList  = xTrailList - trailMargin, xTrailList + trailMargin

                if ( numpy.logical_and(minXTrailList < x1, maxXTrailList > x0).any() ):
                    fpTrailList.append(fp)
                    break

        return fpTrailList


    def maskTrailCandidates(self, exp, aList, bList, detectedBit, trailBit, 
                            trailMargin = 50.0, minTrailArea = 1000
                            ):
        """ select footprints of possible trails and set new mask bit to them 
        """

        mask = exp.getMaskedImage().getMask()
        thresh = afwDetection.Threshold(detectedBit, afwDetection.Threshold.BITMASK)
        fpList = afwDetection.FootprintSet(mask, thresh).getFootprints()

        xmin, ymin = mask.getBBox(afwImage.PARENT).getMin()

        fpTrailList = self.getTrailFootprints(mask, fpList, aList, bList, trailMargin, minTrailArea)
    
        for fp in fpTrailList:
            for s in fp.getSpans():
                x0, x1 = s.getX0() - xmin, s.getX1() - xmin # we may handle patch so need to care offset
                y = s.getY() - ymin
                for x in range(x0, x1):
                    mask.set(x, y, mask.get(x, y) | trailBit)


        # display the resultant mask on ds9
        if True:
            ds9.setMaskPlaneColor('TRAIL', 'cyan')
            ds9.mtv(mask.convertU(), frame = 2)
            ds9.mtv(mask, frame = 3)        

        return fpTrailList


def main():

    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument('inputFile', help = 'input CORR FITS file to mask trails')
    parser.add_argument('outputFile', help = 'output FITS file name')
    parser.add_argument('--min-hough', dest = 'minHough', type = int, default = 7000, help='minimum peak count in Hough phase space to recognize as a peak (default: 7000)')
    parser.add_argument('--min-peak-area', dest = 'minPeakArea', type = int, default = 10, help='minimum connected pixels above the peak count in Hough phase space to recognize as a peak (default: 10)')
    parser.add_argument('--allowed-peak-ratio', dest = 'allowedPeakRatio', type = float, default = 1.5, help='all the Hough peaks should be larger than the first peak devided by this ratio (default: 1.5)')
    parser.add_argument('--min-trail-area', dest = 'minTrailArea', type = int, default = 300, help='minimum area of each trail candiate (default: 300)')
    parser.add_argument('--trail-margin', dest = 'trailMargin', type = float, default = 30.0, help='acceptable spatial separation of actual trails to be masksed from the estimated line features (default: 30.0)')
    parser.add_argument('--bin', dest = 'binSize', type = int, default = 1, help='image binning factor for speed up processing time (default: 1)')
    parser.add_argument('--lsigma', dest = 'lowSigma', type = float, default = None, help='nsigma for image low-threshold for Hough transformation (default: None)')
    parser.add_argument('--hsigma', dest = 'highSigma', type = float, default = None, help='nsigma for image high-threshold for Hough transformation (default: None)')
    parser.add_argument('--min-image', dest = 'minImage', type = float, default = None, help='minimum count for image low-threshold for Hough transformation. This option overrides --lsigma (default: unset)')
    parser.add_argument('--max-image', dest = 'maxImage', type = float, default = None, help='maximum count for image high-threshold for Hough transformation. This option overrides --hsigma (default: unset)')

    args = parser.parse_args()
    inputFile = args.inputFile
    outputFile = args.outputFile

    minHough = args.minHough
    minPeakArea = args.minPeakArea
    allowedPeakRatio = args.allowedPeakRatio
    minTrailArea = args.minTrailArea
    trailMargin = args.trailMargin
    binSize = args.binSize

#    file = 'CORR-0022374-034.fits'
#    file = 'CORR-0013200-090.fits'
#    file = 'CORR-0013200-091.fits'
#    file = 'CORR-0013200-096.fits'
#    file = 'CORR-0013200-097.fits' # many lines are detected
#    file = 'CORR-0013202-098.fits'
#    file = 'CORR-0013202-099.fits'

#    nAngle = 16
#    nRho = 32
###    nAngle = 200
###    nRho = 200


    # so far successful Hough parameters
    nAngle = 500
    nRho = 500
    rMargin = 60

    exp = afwImage.ExposureF(inputFile)
    nxImg, nyImg = exp.getMaskedImage().getImage().getDimensions()
    mask = exp.getMaskedImage().getMask()

    trail = TrailsTask(
        nAngle = nAngle,
        nRho = nRho,
        rMargin = rMargin, 
        minHough = minHough,
        minPeakArea = minPeakArea,
        allowedPeakRatio = allowedPeakRatio,
        minTrailArea = minTrailArea,
        trailMargin = trailMargin,
        binSize = binSize,
        )

    #
    # detect all possible trails
    #
    aList, bList = trail.getTrailCandidates(exp, lowSigma = args.lowSigma, highSigma = args.highSigma, minImage = args.minImage, maxImage = args.maxImage)

    #
    # then, mask possible trails' footprints with 'TRAIL' bit
    #
    mask.addMaskPlane('TRAIL')
    trailBit = mask.getPlaneBitMask('TRAIL')
    detectedBit = mask.getPlaneBitMask('DETECTED')

    fpList = trail.maskTrailCandidates(exp, aList, bList, detectedBit, trailBit, trailMargin = trailMargin, minTrailArea = minTrailArea)
    print 'candidate footprint:', len(fpList)

    if False: # debug
        for fp in fpList:
            print fp.getBBox(), fp.getArea()

    exp.writeFits(outputFile)


if __name__ == '__main__':
    main()

