#!/usr/bin/env python 
import numpy, sys 

import lsst.afw.image as afwImage
import lsst.afw.detection as afwDetection
import lsst.afw.display.ds9 as ds9
import lsst.afw.math as afwMath
import hsc.qa as qa

debugDump=True
debugDisplay=True
debugWrite=True

class TrailsTask(object):

    def __init__(self, **kw):

        if True:
            self.nAngle = None
            self.nRho = None
            self.rMargin1 = None
            self.rMargin2 = None
            self.minHough = None
            self.minHoughScaledByNfp = None
            self.minPeakArea = None
            self.minPeakCounts = None
            self.allowedPeakRatio = None
            self.minTrailArea = None
            self.trailMargin = None
            self.binSize = None
            self.outputFile = 'trail'


        if kw['nAngle']: self.nAngle = kw['nAngle']
        if kw['nRho']: self.nRho = kw['nRho']
        if kw['rMargin1']: self.rMargin1 = kw['rMargin1']
        if kw['rMargin2']: self.rMargin2 = kw['rMargin2']
        if kw['minHough']: self.minHough = kw['minHough']
        if kw['minHoughScaledByNfp']: self.minHoughScaledByNfp = kw['minHoughScaledByNfp']
        if kw['minPeakArea']: self.minPeakArea = kw['minPeakArea']
        if kw['minPeakCounts']: self.minPeakCounts = kw['minPeakCounts']
        if kw['allowedPeakRatio']: self.allowedPeakRatio = kw['allowedPeakRatio']
        if kw['minTrailArea']: self.minTrailArea = kw['minTrailArea']
        if kw['trailMargin']: self.trailMargin = kw['trailMargin']
        if kw['binSize']: self.binSize = kw['binSize']
        if kw['outputFile']: self.outputFile = kw['outputFile']        

    def getTrailCandidates(self, exp, lowSigma = None, highSigma = None, minImage = None, maxImage = None):
        """
        get possible trail candidates based on Hough transformation
        """

        mi = exp.getMaskedImage()
        mask = mi.getMask()

        # binning mode is not extensively tested
        if self.binSize > 1:
            print 'doing binning by: %d x %d' % (binSize, binSize)
            miBinned = afwMath.binImage(mi, self.binSize)
            # sampling due to rho space can be reduced thanks to binning, but angle space keeps the same
            ###himage = self.getHoughImage(miBinned, goodMask = None, badMask = None, 
            himage, self.fpList = self.getHoughImage(miBinned, goodMask = None, badMask = None, 
                                        nRho = int(float(self.nRho) / self.binSize), 
                                        nAngle = self.nAngle, 
                                        lowSigma = lowSigma, highSigma = highSigma, 
                                        minImage = minImage, maxImage = maxImage
                                        )
            himage = afwMath.binImage(himage, self.binSize, 1, afwMath.MEAN) # binning in angle space to use rMargin in the same manner as in the unbinned case
            xHoughPeakList, yHoughPeakList = self.getHoughPeaks(maskBinned, himage, 
                                                                minHough = (self.minHough / self.binSize / self.binSize), 
                                                                allowedPeakRatio = self.allowedPeakRatio, 
                                                                minPeakArea = (self.minPeakArea / self.binSize / self.binSize),
                                                                minPeakCounts = (self.minPeakCounts / self.binSize / self.binSize),
                                                                rMargin1 = (self.rMargin1 / self.binSize),
                                                                rMargin2 = (self.rMargin2 / self.binSize),
                                                                minHoughScaledByNfp = self.minHoughScaledByNfp, 
                                                                )
            # convert xPeak, yPeak into what would be on the original unbinned himage
            xHoughPeakList, yHoughPeakList = xHoughPeakList * float(self.binSize), yHoughPeakList * float(self.binSize)
        else:
            ###himage = self.getHoughImage(mi, goodMask = None, badMask = None, 
            himage, self.fpList = self.getHoughImage(mi, goodMask = None, badMask = None, 
                                        nRho = self.nRho, nAngle = self.nAngle,
                                        lowSigma = lowSigma, highSigma = highSigma, 
                                        minImage = minImage, maxImage = maxImage
                                        )
            xHoughPeakList, yHoughPeakList = self.getHoughPeaks(mask, himage, 
                                                                minHough = self.minHough, 
                                                                allowedPeakRatio = self.allowedPeakRatio, 
                                                                minPeakArea = self.minPeakArea, 
                                                                minPeakCounts = self.minPeakCounts, 
                                                                rMargin1 = self.rMargin1,
                                                                rMargin2 = self.rMargin2,
                                                                minHoughScaledByNfp = self.minHoughScaledByNfp, 
                                                                )

        aList, bList = self.convertHoughToTrails(mask, xHoughPeakList, yHoughPeakList, 
                                                 nRho = self.nRho, nAngle = self.nAngle
                                                 )

        # display the resultant trails on ds9
        if debugDisplay:
            ds9.mtv(exp.getMaskedImage().getImage(), frame = 1) 
            for (a, b) in zip(aList, bList):
                xx = numpy.array([0, mask.getWidth()-1]) 
                yy = a * xx + b 
                points = numpy.array([(x, y) for (x, y) in zip(xx, yy)])
                ds9.line(points, frame = 1, ctype='cyan')

        return aList, bList


    def filterMaskByFlux(self, mask, fpList, badMaskList, badBit, frac = 0.99, fluxSigma = 10.0):
        """
        set bad mask to pixels to be ignored in Hough transformation based on a given flux thresholds, 
        which are specified by flux frac and flux sigma.
        """

        maskTmp = afwImage.MaskU(mask, True)

        #fluxSigma = 50.0  # flux > threshold = (mean + fluxSigma * stdev) is used
        #fluxSigma = 5.0  # flux > threshold = (mean + fluxSigma * stdev) is used
        #---> 5sigma seems to be a bit too small
        #fluxSigma = 10.0  # flux > threshold = (mean + fluxSigma * stdev) is used        
        #frac = 0.99 # treshold should be < (peakCount * frac)
        maskVal = 0
        for maskBit in badMaskList:
            maskVal |= afwImage.MaskU.getPlaneBitMask(maskBit)
        nClip=3.0; nIter=3
        sctrl = afwMath.StatisticsControl(nClip, nIter, maskVal)

        #badBit = maskTmp.getPlaneBitMask('BAD')

        for fp in fpList:

            bbox = fp.getBBox()
            miFp = afwImage.MaskedImageF(mi, bbox)
            stat = afwMath.makeStatistics(miFp, afwMath.MEANCLIP | afwMath.STDEVCLIP | afwMath.MAX, sctrl)
            peakFp = stat.getValue(afwMath.MAX)
            meanFp = stat.getValue(afwMath.MEANCLIP)
            stdevFp = stat.getValue(afwMath.STDEVCLIP)
            threshCount = meanFp + fluxSigma * stdevFp
            fpArea = fp.getArea()
            print '** %s area %f peakFp %f meanFp %f stdevFp %f threshCount %f' % (str(bbox), fpArea, peakFp, meanFp, stdevFp, threshCount),

            if threshCount > peakFp * frac:
                threshCount = peakFp * frac
                print 'newThreshCount %f' % threshCount
            else:
                print ''

            xmin, ymin = maskTmp.getBBox(afwImage.PARENT).getMin()            
            for s in fp.getSpans():
                x0, x1 = s.getX0() - xmin, s.getX1() - xmin # we may handle patch so need to care offset
                y = s.getY() - ymin
                for x in range(x0, x1):
                    if im.get(x, y) < threshCount:
                        maskTmp.set(x, y, maskTmp.get(x, y) | badBit)

        return maskTmp

        

    def filterMaskByHalvingLines(self, mask, fpList, badBit, maxAspectRatio = 50.0, lineWidth = 10):
        """
        set bad mask to pixels to be ignored in Hough transformation, by taking 
        the intersection of x-x & y-y halving lines in each footprint.
        """

        maskTmp = afwImage.MaskU(mask, True)

        xmin, ymin = maskTmp.getBBox(afwImage.PARENT).getMin()
        xmax, ymax = maskTmp.getBBox(afwImage.PARENT).getMax()


        lw = int(lineWidth/2.0)
        for fp in fpList:

            # patchy work to perform getSpans with y-y slice rather than with x-x slice
            bboxFp = fp.getBBox()
            xminFp, xmaxFp = bboxFp.getMinX(), bboxFp.getMaxX()
            yminFp, ymaxFp = bboxFp.getMinY(), bboxFp.getMaxY()
####            print 'minmaxfp:', xminFp, xmaxFp, yminFp, ymaxFp
            nxFp, nyFp = bboxFp.getWidth(), bboxFp.getHeight()

            aspectFp = float(nxFp)/float(nyFp)
            if aspectFp > maxAspectRatio:
                oneAxisOnly = 'y'
            elif aspectFp < (1.0/maxAspectRatio):
                oneAxisOnly = 'x'
            else:
                oneAxisOnly = None

            fpArr = numpy.array([[0] * nyFp] * nxFp)


            #numpy.array([[0] * 5] * 4)
            #array([[0, 0, 0, 0, 0], # x=0
            #       [0, 0, 0, 0, 0], # x=1
            #       [0, 0, 0, 0, 0], # x=2
            #       [0, 0, 0, 0, 0]])# x=3

            # for x-x slice
            for s in fp.getSpans():
                x0, x1 = s.getMinX() - xmin, s.getMaxX() - xmin 
                y = s.getY() - ymin
                halfPos = int((x0 + x1)/2.0)
                hl0 = halfPos - lw
                hl1 = halfPos + lw
                for x in range(x0, x1 + 1):

                    fpArr[x - xminFp][y - yminFp] = 1

                    if oneAxisOnly != 'y':
                        if x < hl0 or x > hl1:
                            maskTmp.set(x, y, maskTmp.get(x, y) | badBit)

            # for y-y slice
            if oneAxisOnly != 'x':
                for x, yline in enumerate(fpArr):
                    ####print x, x+xminFp, len(yline), len(yline)-1+yminFp
                    x += xminFp
                    y0 = -1
                    y1 = -1
                    for y, v in enumerate(yline):
                        if y0 < 0 and v > 0:
                            y0 = y + yminFp
                            continue
                        if y1 < 0 and y0 >= 0 and v == 0:
                            y1 = y - 1 + yminFp
                            break
                    if y1 < 0:
                        y1 = len(yline) - 1 + yminFp
                    ###print '*** (x: y0, y1) = (%d: %d, %d)' % (x, y0, y1)
                    halfPos = int((y0 + y1)/2.0)
                    hl0 = halfPos - lw
                    hl1 = halfPos + lw
                    for y in range(y0, y1 + 1):
                        if (y < hl0 or y > hl1): # and y <= ymax:
                            maskTmp.set(x, y, maskTmp.get(x, y) | badBit)

        ###import pudb;pudb.set_trace()

        return maskTmp
        

    def getHoughImage(self, mi, goodMask = None, badMask = None, nRho = 500, nAngle = 500, 
                      lowSigma = None, highSigma = None, minImage = None, maxImage = None
                      ):
        """
        perform Hough transformation and get Hough image
        """

        mask = mi.getMask()

        # making mask image have only a given faction of the brightest pixels in each footprint
        im = mi.getImage()

        detectedBit = mask.getPlaneBitMask('DETECTED')
        thresh = afwDetection.Threshold(detectedBit, afwDetection.Threshold.BITMASK)
        fpList = afwDetection.FootprintSet(mask, thresh).getFootprints()

        xmin, ymin = mask.getBBox(afwImage.PARENT).getMin()

#        fpTrailList = self.getTrailFootprints(mask, fpList, aList, bList, trailMargin, minTrailArea)


        badMaskList = ['NO_DATA', 'EDGE', 'BAD', 'SAT', 'CR']
        ##badMaskList = ['NO_DATA', 'EDGE', 'BAD', 'SAT']
        ##badMaskList = ['NO_DATA', 'EDGE', 'BAD']

        badBit = mask.getPlaneBitMask('BAD')

        if False:
            maskTmp = self.filterMaskByFlux(mask, fpList, badMaskList, badBit, fluxFrac = 0.99, fluxSigma = 10.0)
        
#        maskTmp = self.filterMaskByHalvingLines(mask, fpList, badBit, maxAspectRatio = 50.0, lineWidth = 5)
        maskTmp = self.filterMaskByHalvingLines(mask, fpList, badBit, maxAspectRatio = 50.0, lineWidth = 10.0)

        if goodMask is None:
            goodMask = maskTmp.getPlaneBitMask("DETECTED")
        if badMask is None:
            badMaskList = ['NO_DATA', 'EDGE', 'BAD', 'SAT', 'CR']
            ##badMaskList = ['NO_DATA', 'EDGE', 'BAD', 'SAT']
            ##badMaskList = ['NO_DATA', 'EDGE', 'BAD']
            #badMaskList = ['NO_DATA', 'EDGE', 'BAD', 'DETECTED']
            #badMaskList = ["NO_DATA", "EDGE", "SAT", "SUSPECT", "INTRP", "BAD"]
            badMask = maskTmp.getPlaneBitMask(badMaskList)

        print 'goodMaskBit = %d, badMaskBit = %d' % (goodMask, badMask)


        # calculate skysigma and threshold for pixels used in Hough
        if minImage is None or maxImage is None:
            maskVal = 0
            for maskBit in badMaskList:
                maskVal |= afwImage.MaskU.getPlaneBitMask(maskBit)
            nClip = 3.0; nIter = 3
            sctrl = afwMath.StatisticsControl(nClip, nIter, maskVal)
            stat = afwMath.makeStatistics(mi, afwMath.MEANCLIP | afwMath.STDEVCLIP | afwMath.MAX, sctrl)
            bkg = stat.getValue(afwMath.MEANCLIP)
            bkgSigma = stat.getValue(afwMath.STDEVCLIP)

            if minImage is None:
                if lowSigma is not None and lowSigma > 0:
                    minImage = bkg + lowSigma * bkgSigma
            if maxImage is None:
                if highSigma is not None and highSigma > 0:
                    maxImage = bkg + highSigma * bkgSigma
                else:
                    maxImage = stat.getValue(afwMath.MAX)

        else: # use the given minImage and maxImage
            pass
    
        #
        # perform Hough transformation
        #
        maskArr = maskTmp.getArray()
        maskArr = numpy.array(maskArr, dtype=numpy.int32)
        maskHough = afwImage.ImageI(maskArr)


        if debugDisplay:
            ds9.mtv(maskHough, frame = 2)
        if debugWrite:
            maskHough.writeFits(self.outputFile + '_mask4hough.fits')

        print 'nAngle:', nAngle, 'nRho:', nRho
        hough = qa.Hough(maskHough, nAngle, nRho)

        if minImage is not None:
            print 'image count threshold applied to Hough transform: (%s, %s) (count)' % (minImage, maxImage)
            im = mi.getImage()
            himage, ang, dis = hough.calcHough(im, minImage, maxImage, goodMask, badMask)
        else:
            print 'image count threshold not applied to Hough transform'
            himage, ang, dis = hough.calcHough(goodMask, badMask)

        print 'Npix effective: ', hough.getNpixEffective()

        self.himage = himage

        # This is the highest peak returned by in Hough class 
        # But we need to detect all possible peaks corresponding to trails, so do not simply use this
        a = -1.0 * (numpy.cos(ang)/numpy.sin(ang))
        b = dis / numpy.sin(ang)
        ####    print ang, dis
        ####    print "y = %g * x + %g" % (a, b)

        if False: ### for debug
            return self.himage
        else:
            return self.himage, fpList
    

    def getHoughPeaks(self, mask, himage, minHough, allowedPeakRatio = None,
                      minPeakArea = 10,  minPeakCounts = 10000, rMargin1 = 10.0, rMargin2 = 60.0, 
                      minHoughScaledByNfp = None, 
                      ):
        """
        select all possible peaks (x, y) in resultant Hough-transformed plane
        """

        # this is a piece of experiments of scaling conditions 
        # based on 2k4k CORR experience to other-size data
        if False:
            maskArr = mask.getArray()
            maskEff = maskArr[ 
                numpy.logical_and((maskArr & mask.getPlaneBitMask("NO_DATA") == 0), 
                                  (maskArr & mask.getPlaneBitMask("EDGE") == 0)) 
                ]

            scale = float(len(maskEff)) / (2048.*4176.)
            if scale < 1: # if the effective area is smaller than 2k4k, we will use 1 as the lowest scale
                scale = 1.0
            minHough *= scale # scaled by effective image area
            print 'effective area: %d (pix) / (2048*4176) = %8.4f : scale_used = %8.4f' % (len(maskEff), float(len(maskEff)) / (2048.*4176.), scale)

        # for experiments of scaling conditions by number of footprints
        nfp = float(len(self.fpList))

        vHough = himage.getArray()
        nyHough, nxHough = vHough.shape
        print 'Hough nx: %d ny: %d' % (nxHough, nyHough)

        # convert hough to 1d array 
        vHough1d = numpy.reshape(vHough, nxHough * nyHough )
        x1dInd = numpy.array( [range(nxHough)] * nyHough, dtype=numpy.int32 )
        y1dInd = numpy.array( [ [y] * nxHough for y in range(nyHough) ], dtype=numpy.int32 )
        x1dInd = numpy.reshape(x1dInd, nxHough * nyHough )
        y1dInd = numpy.reshape(y1dInd, nxHough * nyHough )

        # pick out only high hough pixels
        nCand = len(vHough1d)
        print 'candidates before filtering:', nCand
        if minHoughScaledByNfp:
            indHigh = numpy.logical_and(vHough1d > int(nfp*minHoughScaledByNfp), y1dInd != 500) # clipping 500 is workaround
        else:
            indHigh = numpy.logical_and(vHough1d > minHough, y1dInd != 500) # clipping 500 is workaround
        vHough1dHigh = vHough1d[indHigh]
        x1dIndHigh = x1dInd[indHigh]
        y1dIndHigh = y1dInd[indHigh]

        # grouping high hough-pixel knots, 
        #   assigning one pixel for each knot group by taking the highest hough value in a knot
        xCandList = numpy.array([])
        yCandList = numpy.array([])

        if len(vHough1dHigh) == 0:
            print 'candidates after filtering:', 0
            print 'No trail candidate found'
            # display the resultant Hough peaks on ds9
            if debugDisplay:
                ds9.mtv(himage, frame = 0)
            if debugWrite:
                himage.writeFits(self.outputFile + '_hough.fits')

            return xCandList, yCandList 

        ind1dHighSorted = numpy.argsort(vHough1dHigh)
        #   revert to make it descending
        ind1dHighSorted = ind1dHighSorted[-1::-1]
        vHough1dHighSorted = vHough1dHigh[ind1dHighSorted]
        x1dIndHighSorted = x1dIndHigh[ind1dHighSorted]
        y1dIndHighSorted = y1dIndHigh[ind1dHighSorted]

        # further limit pixels by the hough's peak ratio
        maxPeak = vHough1dHighSorted[0]
        if allowedPeakRatio:
            minPeak = maxPeak / allowedPeakRatio
            print 'allowedPeakRatio = %f' %  allowedPeakRatio
        else:
            minPeak = 0
            print 'allowedPeakRatio unset'
        print 'maxPeak = %f' % maxPeak
        print 'minPeak = %f' % minPeak
        vHough1dFiltered = vHough1dHighSorted
        x1dIndFiltered = x1dIndHighSorted
        y1dIndFiltered = y1dIndHighSorted

        #   start grouping knots, with descending-sorted vHoughArr
        vHoughArr = numpy.copy(vHough1dFiltered)
        xArr = numpy.copy(x1dIndFiltered)
        yArr = numpy.copy(y1dIndFiltered)

        while len(vHoughArr) > 0:

            if debugDump:
                pass
                print '+' * 30
                for x, y, v in zip(xArr, yArr, vHoughArr):
                    print x, y, v, v/nfp
                print '+' * 30

            peakThisKnot = vHoughArr[0] # highest peak in this group
            x0 = xArr[0]
            y0 = yArr[0]
            
            # count up hough points which belongs to this knot group, and
            # carry over the other groups to the next loop
            nPixelsThisKnot = 0
            countsThisKnot = 0
            indOthers = numpy.array([], dtype = int)
            for i, (x, y, v) in enumerate(zip(xArr, yArr, vHoughArr)):
                dist = numpy.hypot(x - x0, y - y0)
                if dist < rMargin2:
                    if dist < rMargin1:
                        nPixelsThisKnot += 1
                        countsThisKnot += v
                else:
                    indOthers = numpy.append(indOthers, i)

            # decide if this group is a significant trail
###            if (peakThisKnot >= minPeak) and (nPixelsThisKnot >= minPeakArea) and (countsThisKnot >= minPeakCounts):
            if (nPixelsThisKnot >= minPeakArea): # minpeak is already screened by minhoughscaled
                xCandList = numpy.append(xCandList, x0)
                yCandList = numpy.append(yCandList, y0)
###                print 'thisKnot(%s, %s)(peak, npixel, counts): (%s, %s, %s) -- selected' % (x0, y0, peakThisKnot, nPixelsThisKnot, countsThisKnot)
                print 'thisKnot(%s, %s)(peak, npixel, counts): (%s, %s, %s) nfp=%d (%s, %s, %s) -- selected' % (x0, y0, peakThisKnot, nPixelsThisKnot, countsThisKnot, nfp, peakThisKnot/nfp, nPixelsThisKnot/nfp, countsThisKnot/nfp)
            else:
                if debugDump:
###                    print 'thisTestedKnot(%s, %s)(peak, npixel, counts): (%s, %s, %s) -- rejected' % (x0, y0, peakThisKnot, nPixelsThisKnot, countsThisKnot)
                    print 'thisTestedKnot(%s, %s)(peak, npixel, counts): (%s, %s, %s) nfp=%d (%s, %s, %s) -- rejected' % (x0, y0, peakThisKnot, nPixelsThisKnot, countsThisKnot, nfp, peakThisKnot/nfp, nPixelsThisKnot/nfp, countsThisKnot/nfp)

            vHoughArr = vHoughArr[indOthers]
            xArr = xArr[indOthers]
            yArr = yArr[indOthers]

        nCandHigh = len(xCandList)
        print 'candidates after filtering:', nCandHigh


        # display the resultant Hough peaks on ds9
        if debugDisplay:
            ds9.mtv(himage, frame = 0)
            for (t, r) in zip(xCandList, yCandList):
                ds9.dot('o', t, r, frame = 0, size = rMargin2, ctype = 'cyan')
        if debugWrite:
            himage.writeFits(self.outputFile + '_hough.fits')

        return xCandList, yCandList


    def convertHoughToTrails(self, mask, xHoughPeakList, yHoughPeakList, nRho, nAngle):
        """
        convert input Hough peaks into line parameters (a:slope, b:intersect)
        """
    
        nxImage, nyImage = mask.getDimensions()

        anglePeakList = numpy.pi * xHoughPeakList / nAngle
        rhoPeakList = (yHoughPeakList - float(nRho)) * int(numpy.hypot(nxImage-1, nyImage-1) + 1) / float(nRho)
        
        aList = -1.0 * (numpy.cos(anglePeakList)/numpy.sin(anglePeakList))
        bList = rhoPeakList / numpy.sin(anglePeakList)

        # !!!
        # better to handle cases of angle = 0 or a = inf

        return aList, bList


    def getTrailFootprints(self, mask, fpList, aList, bList, trailMargin, minTrailArea):
        """ select footprints of trail candidates from the input footprint list
        """

# check if this footprint overlaps with the trails candidates in each y 

        xmin, ymin = mask.getBBox(afwImage.PARENT).getMin()

        fpTrailList = []
        for fp in fpList:

            # footprint should be larger than the given minArea to be a trail
            if False:
                print '*** minTrailArea:', fp.getArea()
            if fp.getArea() < minTrailArea:
                continue

            # check if this fp overlaps with trail candidates
            for s in fp.getSpans():
                x0, x1 = s.getX0() -xmin, s.getX1() -xmin
                y = s.getY() - ymin

                # define trail regions with margin for recognizing this footprint as a candidate
                xTrailList = (y - bList) / aList
                minXTrailList, maxXTrailList  = xTrailList - trailMargin, xTrailList + trailMargin

                if ( numpy.logical_and(minXTrailList < x1, maxXTrailList > x0).any() ):
                    fpTrailList.append(fp)
                    break

        return fpTrailList


    def maskTrailCandidates(self, exp, aList, bList, detectedBit, trailBit, 
                            trailMargin = 50.0, minTrailArea = 1000
                            ):
        """ select footprints of possible trails and set new mask bit to them 
        """

        mask = exp.getMaskedImage().getMask()
        thresh = afwDetection.Threshold(detectedBit, afwDetection.Threshold.BITMASK)
        fpList = afwDetection.FootprintSet(mask, thresh).getFootprints()

        xmin, ymin = mask.getBBox(afwImage.PARENT).getMin()

        fpTrailList = self.getTrailFootprints(mask, fpList, aList, bList, trailMargin, minTrailArea)
    
        for fp in fpTrailList:
            for s in fp.getSpans():
                x0, x1 = s.getX0() - xmin, s.getX1() - xmin # we may handle patch so need to care offset
                y = s.getY() - ymin
                for x in range(x0, x1):
                    mask.set(x, y, mask.get(x, y) | trailBit)


        # display the resultant mask on ds9
        if debugDisplay:
            ds9.setMaskPlaneColor('TRAIL', 'cyan')
            ds9.mtv(mask.convertU(), frame = 3)

        return fpTrailList


def main():

    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument('inputFile', help = 'input CORR FITS file to mask trails')
    parser.add_argument('outputFile', help = 'prefix of output FITS file name')
    parser.add_argument('--min-hough', dest = 'minHough', type = int, default = 1000, help='minimum peak count in Hough phase space to recognize as a peak. if --do-scale-minhoug is set, this option is ignored (default: 1000)')
    parser.add_argument('--min-peak-area', dest = 'minPeakArea', type = int, default = 10, help='minimum connected pixels above the peak count in Hough phase space to recognize as a peak (default: 10)')
    parser.add_argument('--min-peak-counts', dest = 'minPeakCounts', type = int, default = 10000, help='minimum sum of peak counts over the pixels within peak-area-radius1 above the min-hough count in Hough phase space to recognize as a peak (default: 10000)')
    parser.add_argument('--peak-area-radius1', dest = 'rMargin1', type = float, default = 10.0, help='inner margin radius to count up peak pixels in hough phase space (default: 10.0)')
    parser.add_argument('--peak-area-radius2', dest = 'rMargin2', type = float, default = 100.0, help='outer margin radius for grouping individual peak pixels in hough phase space. pixels within this radius belongs to the same peak and will not go to another neighboring peak  (default: 100.0)')
    parser.add_argument('--allowed-peak-ratio', dest = 'allowedPeakRatio', type = float, default = None, help='all the Hough peaks should be larger than the first peak devided by this ratio (default: unset)')
    parser.add_argument('--min-trail-area', dest = 'minTrailArea', type = int, default = 300, help='minimum area of each trail candiate (default: 300)')
    parser.add_argument('--trail-margin', dest = 'trailMargin', type = float, default = 30.0, help='acceptable spatial separation of actual trails to be masksed from the estimated line features (default: 30.0)')
    parser.add_argument('--bin', dest = 'binSize', type = int, default = 1, help='image binning factor for speed up processing time (default: 1)')
    parser.add_argument('--lsigma', dest = 'lowSigma', type = float, default = None, help='nsigma for image low-threshold for Hough transformation (default: None)')
    parser.add_argument('--hsigma', dest = 'highSigma', type = float, default = None, help='nsigma for image high-threshold for Hough transformation (default: None)')
    parser.add_argument('--min-image', dest = 'minImage', type = float, default = None, help='minimum count for image low-threshold for Hough transformation. This option overrides --lsigma (default: unset)')
    parser.add_argument('--max-image', dest = 'maxImage', type = float, default = None, help='maximum count for image high-threshold for Hough transformation. This option overrides --hsigma (default: unset)')
    parser.add_argument('--min-hough-scaled', dest = 'minHoughScaledByNfp', default = None, help = 'min-hough if given, --min-hough will be ignored.')

    args = parser.parse_args()
    inputFile = args.inputFile
    outputFile = args.outputFile

    minHough = args.minHough
    minPeakArea = args.minPeakArea
    minPeakCounts = args.minPeakCounts
    allowedPeakRatio = args.allowedPeakRatio
    minTrailArea = args.minTrailArea
    trailMargin = args.trailMargin
    binSize = args.binSize
    rMargin1 = args.rMargin1
    rMargin2 = args.rMargin2

    minHoughScaledByNfp = args.minHoughScaledByNfp

#    file = 'CORR-0022374-034.fits'
#    file = 'CORR-0013200-090.fits'
#    file = 'CORR-0013200-091.fits'
#    file = 'CORR-0013200-096.fits'
#    file = 'CORR-0013200-097.fits' # many lines are detected
#    file = 'CORR-0013202-098.fits'
#    file = 'CORR-0013202-099.fits'

#    nAngle = 16
#    nRho = 32
###    nAngle = 200
###    nRho = 200


    # so far successful Hough parameters
    nAngle = 500
    nRho = 500

    exp = afwImage.ExposureF(inputFile)
    nxImg, nyImg = exp.getMaskedImage().getImage().getDimensions()
    mask = exp.getMaskedImage().getMask()

    trail = TrailsTask(
        nAngle = nAngle,
        nRho = nRho,
        rMargin1 = rMargin1, 
        rMargin2 = rMargin2, 
        minHough = minHough,
        minHoughScaledByNfp = minHoughScaledByNfp, 
        minPeakArea = minPeakArea,
        minPeakCounts = minPeakCounts,
        allowedPeakRatio = allowedPeakRatio,
        minTrailArea = minTrailArea,
        trailMargin = trailMargin,
        binSize = binSize,
        outputFile = outputFile, 
        )

    #
    # detect all possible trails
    #
    aList, bList = trail.getTrailCandidates(exp, lowSigma = args.lowSigma, highSigma = args.highSigma, minImage = args.minImage, maxImage = args.maxImage)

    #
    # then, mask possible trails' footprints with 'TRAIL' bit
    #
    mask.addMaskPlane('TRAIL')
    trailBit = mask.getPlaneBitMask('TRAIL')
    detectedBit = mask.getPlaneBitMask('DETECTED')

    fpList = trail.maskTrailCandidates(exp, aList, bList, detectedBit, trailBit, trailMargin = trailMargin, minTrailArea = minTrailArea)
    print 'candidate footprint:', len(fpList)

    if False: # debug
        for fp in fpList:
            print fp.getBBox(), fp.getArea()

    exp.writeFits(outputFile + '_exp.fits')


if __name__ == '__main__':
    main()

