#!/usr/bin/env python 
import numpy, sys 

from lsst.pex.config import Config, Field, ConfigurableField, ListField, DictField, ConfigField
import lsst.pipe.base as pipeBase

import lsst.afw.image as afwImage
import lsst.afw.detection as afwDetection
import lsst.afw.display.ds9 as ds9
import lsst.afw.math as afwMath
import hsc.qa as qa


class HscTrailConfig(Config):
    """ 
    Config class for detecting satellite trails
    """

    rMargin1 = Field(dtype = float, default = 10.0, 
                     doc = 'radius to sum up Hough peak pixels in a peak group')
    rMargin2 = Field(dtype = float, default = 60.0, # 100?
                     doc = 'Hough peaks outside this radius will belong to other peak groups')
    nAngle = Field(dtype = int, default = 500, 
                   doc = 'number of sampling of tangential angle of trails in the Hough transformation')
    nRho = Field(dtype = int, default = 500, 
                 doc = 'number of sampling of radial distance to trails in the Hough transformation')
    minHough = Field(dtype = int, default = 1000, 
                     doc = 'minimum peak count in Hough phase space to recognize as a peak. if --do-scale-minhoug is set, this option is ignored')
    minPeakArea = Field(dtype = int, default = 2, 
                        doc = 'minimum connected pixels above the peak count in Hough phase space to recognize as a peak')
    minPeakCounts = Field(dtype = int, default = -1, 
                          doc = 'minimum sum of peak counts over the pixels within peak-area-radius1 above the min-hough count in Hough phase space to recognize as a peak')

    allowedPeakRatio = Field(dtype = float, default = -1.0,
                             doc = 'all the Hough peaks should be larger than the first peak devided by this ratio')
    minTrailArea = Field(dtype = int, default = 300, 
                         doc = 'minimum area of footprint to be regcognized as a trail candiate')
    trailMargin = Field(dtype = float, default = 30.0, 
                        doc = 'largest spatial separation of actual trails from the estimated line to be masksed')
    binSize = Field(dtype = int, default = 1, 
                    doc = 'image binning factor for speed up processing time')
    lowSigma = Field(dtype = float, default = 2.0,
                     doc = 'nsigma for image low-threshold to be included in Hough transformation. negative value is no effective.')
    highSigma = Field(dtype = float, default = -1.0, 
                      doc = 'nsigma for image high-threshold to be included in Hough transformation. negative value is no effective.')
    minImageValue = Field(dtype = float, default = -1.0,
                     doc = 'minimum count of image to be counted in Hough transformation. If set to positive, this option overrides lowSigma')
    maxImageValue = Field(dtype = float, default = -1.0,
                     doc = 'maximum count for image to be counted in Hough transformation. If set to positive, this option overrides highSigma')
    minHoughPerFootprint = Field(dtype = int, default = -1,
                                 doc = 'min-hough per footprint. if this is set to > 0, minHough will be ignored.')
    doNarrowFootprints = Field(dtype = bool, default = True, doc = 'squash footprints before hough transformation?')
    narrowedLineWidth = Field(dtype = float, default = 10.0, 
                              doc = 'line width (px) after narrowing footprints as a preprocess of hough transformation.') # 10 seems to catch more falses

    doDropRoundFootprints = Field(dtype = bool, default = True, doc = "drop footprints with relatively round sources before hough transformation? When doDropRoundFootprints is True, a footprint satisfying the both conditions is removed in hough transformation: 1) bbox's aspect ratio (x/y) is larger than minFootprintAspectRatio, and 2) a fraction of the detected area of the footprint is larger than maxDetToFpRatio.")
    minFootprintAspectRatio = Field(dtype = float, default = 1.5, doc = "minimum aspect ratio (x/y) of a footprint not to be excluded in drop round footprint.")
    maxFootprintAspectRatio = Field(dtype = float, default = 5.0, doc = "maximum aspect ratio (x/y) of each footprint to be narrowed in both x and y directions.")
    maxDetToBBoxRatio = Field(dtype = float, default = 0.35, doc = "maximum ratio of detected area and footprint bbox area not to be excluded in drop round footprint.")
    maxNspansPerTrail = Field(dtype = int, default = 4, doc = "maximum number of spans anticipated in a trail footprint.")
    outputFile = Field(dtype = str, default = 'test', doc = 'prefix of output file name')
    doWriteFile = Field(dtype = bool, default = False, doc = 'write intermediate output files')

    def setDefaults(self):
        super(HscTrailConfig, self).setDefaults()


class HscTrailTask(pipeBase.Task):

    """
    A task to detect and mask satellite trails.
    """

    _DefaultName = 'trail'
    ConfigClass = HscTrailConfig

    def __init__(self, *args, **kwargs):
        super(HscTrailTask, self).__init__(*args, **kwargs)


    def run(self, exp, display = False, debug = True):
        """
        detect and mask candidates of satellite trails.
        """

        nxImg, nyImg = exp.getMaskedImage().getImage().getDimensions()
        mask = exp.getMaskedImage().getMask()

        #
        # detect all possible trails
        #
        aList, bList = self.getTrailCandidates(exp, lowSigma = self.config.lowSigma, highSigma = self.config.highSigma, minImageValue = self.config.minImageValue, maxImageValue = self.config.maxImageValue, display = display, debug = debug)

        #
        # then, mask possible trails' footprints with 'TRAIL' bit
        #
        if False: # we do not increase mask plane now to cope with the current version of bit mask
            mask.addMaskPlane('TRAIL')
            trailBit = mask.getPlaneBitMask('TRAIL') | mask.getPlaneBitMask('BAD')
        trailBit = mask.getPlaneBitMask('BAD')
        detectedBit = mask.getPlaneBitMask('DETECTED')

        fpList = self.maskTrailCandidates(exp, aList, bList, detectedBit, trailBit, trailMargin = self.config.trailMargin, minTrailArea = self.config.minTrailArea, display = display, debug = debug)

        if debug:
            fpListTmp = numpy.array(fpList)
            fpAreaList = numpy.array([])
            for fp in fpListTmp:
                fpAreaList = numpy.append(fpAreaList, fp.getArea())
            ind = numpy.argsort(fpAreaList)
            fpListTmp = fpListTmp[ind]
                
            for i, fp in enumerate(fpListTmp):
                bbox = fp.getBBox()
                bboxArea = bbox.getArea()
                fpArea = fp.getArea()
                nxFp, nyFp = bbox.getWidth(), bbox.getHeight()
                bboxCenX, bboxCenY = (bbox.getMinX() + bbox.getMaxX())/2., (bbox.getMinY() + bbox.getMaxY())/2.
                self.log.info('debug: fp %d: area: %s bbox area %s at (%s %s) size (%s %s) aratio (x/y, y/x): (%s , %s) det/bbox: %s' % (i, fpArea, bboxArea, bboxCenX, bboxCenY, nxFp, nyFp, float(nxFp)/nyFp, float(nyFp)/nxFp, float(fpArea)/bboxArea))

        self.log.info('number of trail candidate footprints masked: %s' % len(fpList))

        if False: # debug
            for fp in fpList:
                print fp.getBBox(), fp.getArea()

        return

    def getTrailCandidates(self, exp, lowSigma = None, highSigma = None, minImageValue = None, maxImageValue = None, display = False, debug = False):
        """
        get possible trail candidates based on Hough transformation
        """

        mi = exp.getMaskedImage()
        mask = mi.getMask()

        # binning mode is not extensively tested
        if self.config.binSize > 1:
            self.log.info('binning by: %d x %d' % (binSize, binSize))
            miBinned = afwMath.binImage(mi, self.config.binSize)
            # sampling due to rho space can be reduced thanks to binning, but angle space keeps the same
            ###himage = self.getHoughImage(miBinned, goodMask = None, badMask = None, 
            himage, self.fpList = self.getHoughImage(miBinned, goodMask = None, badMask = None, 
                                        nRho = int(float(self.config.nRho) / self.config.binSize), 
                                        nAngle = self.config.nAngle, 
                                        lowSigma = lowSigma, highSigma = highSigma, 
                                        minImageValue = minImageValue, maxImageValue = maxImageValue
                                        )
            himage = afwMath.binImage(himage, self.config.binSize, 1, afwMath.MEAN) # binning in angle space to use rMargin in the same manner as in the unbinned case
            xHoughPeakList, yHoughPeakList = \
                self.getHoughPeaks(maskBinned, himage, 
                                   minHough = (self.config.minHough / self.config.binSize / self.config.binSize), 
                                   allowedPeakRatio = self.config.allowedPeakRatio, 
                                   minPeakArea = (self.config.minPeakArea / self.config.binSize / self.config.binSize),
                                   minPeakCounts = (self.config.minPeakCounts / self.config.binSize / self.config.binSize),
                                   rMargin1 = (self.config.rMargin1 / self.config.binSize),
                                   rMargin2 = (self.config.rMargin2 / self.config.binSize),
                                   minHoughPerFootprint = self.config.minHoughPerFootprint,
                                   debug = debug,
                                   )
            # convert xPeak, yPeak into what would be on the original unbinned himage
            xHoughPeakList, yHoughPeakList = xHoughPeakList * float(self.config.binSize), yHoughPeakList * float(self.config.binSize)
        else:
            ###himage = self.getHoughImage(mi, goodMask = None, badMask = None, 
            himage, self.fpList = self.getHoughImage(mi, goodMask = None, badMask = None, 
                                        nRho = self.config.nRho, nAngle = self.config.nAngle,
                                        lowSigma = lowSigma, highSigma = highSigma, 
                                        minImageValue = minImageValue, maxImageValue = maxImageValue,
                                        display = display, debug = debug, 
                                        )
            xHoughPeakList, yHoughPeakList = self.getHoughPeaks(mask, himage, 
                                                                minHough = self.config.minHough, 
                                                                allowedPeakRatio = self.config.allowedPeakRatio, 
                                                                minPeakArea = self.config.minPeakArea, 
                                                                minPeakCounts = self.config.minPeakCounts, 
                                                                rMargin1 = self.config.rMargin1,
                                                                rMargin2 = self.config.rMargin2,
                                                                minHoughPerFootprint = self.config.minHoughPerFootprint, 
                                                                display = display,
                                                                )

        aList, bList = self.convertHoughToTrails(mask, xHoughPeakList, yHoughPeakList, 
                                                 nRho = self.config.nRho, nAngle = self.config.nAngle
                                                 )

        # display the resultant trails on ds9
        if display:
            ds9.mtv(exp.getMaskedImage().getImage(), frame = 5) 
            for (a, b) in zip(aList, bList):
                xx = numpy.array([0, mask.getWidth()-1]) 
                yy = a * xx + b 
                points = numpy.array([(x, y) for (x, y) in zip(xx, yy)])
                ds9.line(points, frame = 5, ctype='cyan')

        return aList, bList


    def filterMaskByFlux(self, mask, fpList, badMaskList, badBit, frac = 0.99, fluxSigma = 10.0):
        """
        set bad mask to pixels to be ignored in Hough transformation based on a given flux thresholds, 
        which are specified by flux frac and flux sigma.
        """

        maskTmp = afwImage.MaskU(mask, True)

        #fluxSigma = 50.0  # flux > threshold = (mean + fluxSigma * stdev) is used
        #fluxSigma = 5.0  # flux > threshold = (mean + fluxSigma * stdev) is used
        #---> 5sigma seems to be a bit too small
        #fluxSigma = 10.0  # flux > threshold = (mean + fluxSigma * stdev) is used        
        #frac = 0.99 # treshold should be < (peakCount * frac)
        maskVal = 0
        for maskBit in badMaskList:
            maskVal |= afwImage.MaskU.getPlaneBitMask(maskBit)
        nClip=3.0; nIter=3
        sctrl = afwMath.StatisticsControl(nClip, nIter, maskVal)

        #badBit = maskTmp.getPlaneBitMask('BAD')

        for fp in fpList:

            bbox = fp.getBBox()
            miFp = afwImage.MaskedImageF(mi, bbox)
            stat = afwMath.makeStatistics(miFp, afwMath.MEANCLIP | afwMath.STDEVCLIP | afwMath.MAX, sctrl)
            peakFp = stat.getValue(afwMath.MAX)
            meanFp = stat.getValue(afwMath.MEANCLIP)
            stdevFp = stat.getValue(afwMath.STDEVCLIP)
            threshCount = meanFp + fluxSigma * stdevFp
            fpArea = fp.getArea()
            self.log.info('** %s area %f peakFp %f meanFp %f stdevFp %f threshCount %f' % (str(bbox), fpArea, peakFp, meanFp, stdevFp, threshCount))

            if threshCount > peakFp * frac:
                threshCount = peakFp * frac
                self.log.info('newThreshCount %f' % threshCount)

            xmin, ymin = maskTmp.getBBox(afwImage.PARENT).getMin()            
            for s in fp.getSpans():
                x0, x1 = s.getX0() - xmin, s.getX1() - xmin # we may handle patch so need to care offset
                y = s.getY() - ymin
                for x in range(x0, x1):
                    if im.get(x, y) < threshCount:
                        maskTmp.set(x, y, maskTmp.get(x, y) | badBit)

        return maskTmp


    def filterMaskByNarrowFootprints(self, mask, fpList, badBit, minFootprintAspectRatio = 1.5, maxFootprintAspectRatio = 5.0, lineWidth = 10.0, display = False, debug = False):
        """
        set bad mask to pixels to be ignored in Hough transformation, by taking 
        the intersection of x-x & y-y of footprints and narrowing(squashing) line width
        of each candidate footprint.
        """

        def getMaxNumOfSpansPerY(fp):
            """ return 1) maximum number of spans per y row in a footprint, 
                       2) number of y-lines where num of spans exceeds the maxNspansPerTrail
                as an indicator of a blended source.
                810 452
            """
            scans = fp.getSpans() # assum scans is sorted by y coord
            this_y = -1
            maxNumOfSpans = 0
            nyManySpans = 0
            nspans = 0 

            #nySpans = 0 #debug
            #histNspans = {} #debug

            for s in scans:
                y = s.getY()
                if y == this_y: # belongs to same y
                    nspans += 1
                else: # goes to next y
                    #if histNspans.has_key(nspans): #debug
                    #    histNspans[nspans] += 1 #debug
                    #else: #debug
                    #    histNspans[nspans] = 1 #debug
                    if nspans > self.config.maxNspansPerTrail: 
                        nyManySpans += 1
                    if nspans > maxNumOfSpans:
                        maxNumOfSpans = nspans
                    this_y = y
                    nspans = 1

            #debug
            #for k in sorted(histNspans):
            #    self.log.info('nSpans %s  ny = %s' % (k, histNspans[k]))

            return maxNumOfSpans, nyManySpans
                    
        maskTmp = afwImage.MaskU(mask, True)

        if debug and display:
            maskTmp_x = afwImage.MaskU(maskTmp, True)
            maskTmp_y = afwImage.MaskU(maskTmp, True)
            maskTmp0 = afwImage.MaskU(maskTmp, True)
            goodMaskVal = afwImage.MaskU.getPlaneBitMask("DETECTED")
            badMaskList = ['NO_DATA', 'EDGE', 'BAD', 'SAT', 'INTRP', 'CR']
            ignoreMaskVal = 0
            for maskBit in badMaskList:
                ignoreMaskVal |= afwImage.MaskU.getPlaneBitMask(maskBit)
            dx, dy = maskTmp0.getDimensions()
            for y in range(dy):
                for x in range(dx):
                    maskBit = maskTmp0.get(x, y)
                    isGood = maskBit & goodMaskVal
                    isIgnore = maskBit & ignoreMaskVal
                    maskTmp0.set(x, y, maskBit if (isIgnore==0 and isGood >0) else 0)
            ds9.mtv(maskTmp0.convertU(), frame = 4)
            maskTmp0.convertU().writeFits('det.fits')


        xmin, ymin = maskTmp.getBBox(afwImage.PARENT).getMin()
        xmax, ymax = maskTmp.getBBox(afwImage.PARENT).getMax()

        lw = int(lineWidth/2.0)
        for ifp, fp in enumerate(fpList):

            bboxFp = fp.getBBox()
            xminFp, xmaxFp = bboxFp.getMinX(), bboxFp.getMaxX()
            yminFp, ymaxFp = bboxFp.getMinY(), bboxFp.getMaxY()
            #print 'minmaxfp:', xminFp, xmaxFp, yminFp, ymaxFp
            nxFp, nyFp = bboxFp.getWidth(), bboxFp.getHeight()

            aspectFp = float(nxFp)/float(nyFp)
            fpArea = fp.getArea()
            bboxArea = bboxFp.getArea()
            detToBBoxRatio = float(fpArea) / bboxArea

            """ matrix to apply squash footprints
          
            # of spans    detToBBox              aspectFp(nx/ny)           xy direction for fp squash
            in a fp                                                        (or bad-masked if doDropRoundFootprints)
            <= maxNspans  <=maxDetToBBoxRatio   < minFootprintAspectRatio  x or y (trail-like?) 
             (~4)         <=maxDetToBBoxRatio   >=minFootprintAspectRatio  x or y (likely trail)
                          > maxDetToBBoxRatio   < minFootprintAspectRatio  xy or masked out if doDropRoundFootprints
                          > maxDetToBBoxRatio   >=minFootprintAspectRatio  xy (likely elongated object)
            > maxNspans   <=maxDetToBBoxRatio   < minFootprintAspectRatio  xy (blended object)  
                          <=maxDetToBBoxRatio   >=minFootprintAspectRatio  
                                             and <maxFootprintAspectRatio  xy (blended elongated object)
                          <=maxDetToBBoxRatio   >=maxFootprintAspectRatio  x or y (trail-like)
                          > maxDetToBBoxRatio   < minFootprintAspectRatio  xy or masked out if doDropRoundFootprints
                          > maxDetToBBoxRatio   >=minFootprintAspectRatio  xy (blended elongated object)  
                            (~0.5)                (~1.2)  
            """ 
            # for rejecting relatively-round real sources
            if self.config.doDropRoundFootprints: # drop round sources
                if aspectFp < minFootprintAspectRatio and aspectFp > 1./minFootprintAspectRatio: # and fp.getArea() > 5000:

                    # trails should not show a large fraction of detected pixels out of the footprint bbox
                    if detToBBoxRatio > self.config.maxDetToBBoxRatio: # this may be a non-trail source
                        for s in fp.getSpans():
                            x0, x1 = s.getMinX() - xmin, s.getMaxX() - xmin 
                            y = s.getY() - ymin
                            for x in range(x0, x1 + 1):
                                maskTmp.set(x, y, maskTmp.get(x, y) | badBit)

            # for narrowing (squash) footprints 
            if self.config.doNarrowFootprints:

                # determine directions (x or y) to quash footprints
                maxNumOfSpans, nyManySpans = getMaxNumOfSpansPerY(fp)
                if maxNumOfSpans <= self.config.maxNspansPerTrail: # fp has < only a few spans
                    if detToBBoxRatio <= self.config.maxDetToBBoxRatio:
                        if aspectFp > 1.0: # horizontally long
                            axisToSquash = 'y'
                        else: # vertically long
                            axisToSquash = 'x'
                    else: 
                        if aspectFp < minFootprintAspectRatio and aspectFp > 1./minFootprintAspectRatio and self.config.doDropRoundFootprints: # round souce treated above
                            axisToSquash = None
                        else:
                            axisToSquash = 'xy'

                else: # fp has many spans
                    if detToBBoxRatio > self.config.maxDetToBBoxRatio:
                        if aspectFp < minFootprintAspectRatio and aspectFp > 1./minFootprintAspectRatio:
                            if self.config.doDropRoundFootprints: # round souce treated aboveself.config.doDropRoundFootprints: # round souce treated above
                                axisToSquash = None
                            else:
                                axisToSquash = 'xy'
                        else:
                            axisToSquash = 'xy'

                    else: # sparse footprint
                        if aspectFp < maxFootprintAspectRatio and aspectFp > 1./maxFootprintAspectRatio: # relatively short footprint
                            axisToSquash = 'xy'
                        else: # long footprint
                            if aspectFp > 1.0: # horizontally long
                                axisToSquash = 'y'
                            else: # vertically long
                                axisToSquash = 'x'

                if debug:
                    self.log.info('(ifp, maxNspans, nyManySpans, fpArea, bboxArea, bboxFp, aspectFp, detToBBox): (%s %s %s %s %s %s %s %s)' % (ifp, maxNumOfSpans, nyManySpans, fpArea, bboxArea, bboxFp, aspectFp, detToBBoxRatio))

                if False:
                    if aspectFp > maxFootprintAspectRatio:
                        axisToSquash = 'y'
                    elif aspectFp < (1.0/maxFootprintAspectRatio):
                        axisToSquash = 'x'
                    else:
                        axisToSquash = 'xy'

                fpArr = numpy.array([[0] * nyFp] * nxFp)
                # numpy.array([[0] * 5] * 4)
                # array([[0, 0, 0, 0, 0], # x=0
                #       [0, 0, 0, 0, 0], # x=1
                #       [0, 0, 0, 0, 0], # x=2
                #       [0, 0, 0, 0, 0]])# x=3

                # for x-x squash
                for s in fp.getSpans(): # this block should be run for any footprint
                    x0, x1 = s.getMinX() - xmin, s.getMaxX() - xmin 
                    y = s.getY() - ymin
                    halfPos = int((x0 + x1)/2.0)
                    hl0 = halfPos - lw
                    hl1 = halfPos + lw
                    for x in range(x0, x1 + 1):
                        
                        fpArr[x - xminFp][y - yminFp] = 1

                        if axisToSquash in ('x', 'xy'): # x-x squash
                            if x < hl0 or x > hl1:
                                maskTmp.set(x, y, maskTmp.get(x, y) | badBit)
                                if debug and display:
                                    maskTmp_x.set(x, y, maskTmp_x.get(x, y) | badBit)

                # for y-y squash
                #   get spans in y-y slice based fpArr set in the above section
                if axisToSquash in ('y', 'xy'):

                    #squashFootprintInY(fpArr, mask, xminFp, yminFp)

                    for x, yline in enumerate(fpArr):
                        ####print x, x+xminFp, len(yline), len(yline)-1+yminFp

                        spansY = [] # spans locating in this x

                        x += xminFp
                        y0 = -1
                        y1 = -1
                        for y, v in enumerate(yline):
                            if y0 < 0 and v > 0: # (x, y_min) of footprint
                                y0 = y + yminFp
                                continue
                            if y1 < 0 and y0 >= 0 and v == 0:
                                y1 = y - 1 + yminFp
                                spansY.append([x, [y0, y1]]) # add this span and look for the next span
                                y0 = -1
                                y1 = -1

                        if y0 > 0 and y1 < 0:  # footprint extends to the y edge of the image
                            y1 = len(yline) - 1 + yminFp
                            spansY.append([x, [y0, y1]])

                        if debug:
                            self.log.info('*** fpidx (x: y0, y1) = %d (%d: %d, %d) (nxFp, nyFp) = (%s, %s) ' % (ifp, x, y0, y1, nxFp, nyFp))

                        for x, (y0, y1) in spansY:
                            halfPos = int((y0 + y1)/2.0)
                            hl0 = halfPos - lw
                            hl1 = halfPos + lw
                            for y in range(y0, y1 + 1):
                                if (y < hl0 or y > hl1): # and y <= ymax:
                                    maskTmp.set(x, y, maskTmp.get(x, y) | badBit)
                                    if debug and display:
                                        maskTmp_y.set(x, y, maskTmp_y.get(x, y) | badBit)

        if debug and display:
            badMaskList = ['NO_DATA', 'EDGE', 'BAD', 'SAT', 'INTRP', 'CR']
            #badMaskList = ['BAD']
            ignoreMaskVal = 0
            for maskBit in badMaskList:
                ignoreMaskVal |= afwImage.MaskU.getPlaneBitMask(maskBit)
            goodMaskVal = afwImage.MaskU.getPlaneBitMask("DETECTED")

            maskTmp2x = afwImage.MaskU(maskTmp_x, True)
            maskTmp2y = afwImage.MaskU(maskTmp_y, True)
            maskTmp2 = afwImage.MaskU(maskTmp, True)
            dx, dy = maskTmp.getDimensions()
            for y in range(dy):
                for x in range(dx):
                    maskBit = maskTmp2x.get(x, y)
                    isIgnore = maskBit & ignoreMaskVal
                    isGood = maskBit & goodMaskVal
                    maskTmp2x.set(x, y, maskBit if (isIgnore==0 and isGood >0) else 0)
            for y in range(dy):
                for x in range(dx):
                    maskBit = maskTmp2y.get(x, y)
                    isIgnore = maskBit & ignoreMaskVal
                    isGood = maskBit & goodMaskVal
                    maskTmp2y.set(x, y, maskBit if (isIgnore==0 and isGood >0) else 0)
            for y in range(dy):
                for x in range(dx):
                    maskBit = maskTmp2.get(x, y)
                    isIgnore = maskBit & ignoreMaskVal
                    isGood = maskBit & goodMaskVal
                    maskTmp2.set(x, y, maskBit if (isIgnore==0 and isGood >0) else 0)

            maskTmp2x.convertU().writeFits('x.fits')
            maskTmp2y.convertU().writeFits('y.fits')
            maskTmp2.convertU().writeFits('xy.fits')

        return maskTmp
        

    def getHoughImage(self, mi, goodMask = None, badMask = None, nRho = 500, nAngle = 500, 
                      lowSigma = None, highSigma = None, minImageValue = None, maxImageValue = None,
                      display = False, debug = False,
                      ):
        """
        perform Hough transformation and get Hough image
        """

        mask0 = mi.getMask()
        mask = afwImage.MaskU(mask0, True) # have a deepcopy to manipulate the mask bits only for hough transform

        nxMask, nyMask = mask.getDimensions()
        #ignoreBit = mask.getPlaneBitMask('UNMASKEDNAN') | mask.getPlaneBitMask('EDGE')
        # trim CCD edges
        for x in range(15): # left edge
            for y in range(nyMask):
                mask.set(x, y, 0)
        for x in range(nxMask - 15, nxMask): # right edge
            for y in range(nyMask):
                mask.set(x, y, 0)
        for y in range(15): # bottom edge
            for x in range(nxMask):
                mask.set(x, y, 0)
        for y in range(nyMask - 15, nyMask): # top edge
            for x in range(nxMask):
                mask.set(x, y, 0)

        # making mask image have only a given faction of the brightest pixels in each footprint
        im = mi.getImage()

        detectedBit = mask.getPlaneBitMask('DETECTED')
        thresh = afwDetection.Threshold(detectedBit, afwDetection.Threshold.BITMASK)
        fpList = afwDetection.FootprintSet(mask, thresh).getFootprints()

        xmin, ymin = mask.getBBox(afwImage.PARENT).getMin()

#        fpTrailList = self.getTrailFootprints(mask, fpList, aList, bList, trailMargin, minTrailArea)


        badMaskList = ['NO_DATA', 'EDGE', 'BAD', 'SAT', 'CR']
        badBit = mask.getPlaneBitMask('BAD')

        if False:
            maskTmp = self.filterMaskByFlux(mask, fpList, badMaskList, badBit, fluxFrac = 0.99, fluxSigma = 10.0)
        
        maskTmp = self.filterMaskByNarrowFootprints(mask, fpList, badBit, 
                                               minFootprintAspectRatio = self.config.minFootprintAspectRatio, 
                                               maxFootprintAspectRatio = self.config.maxFootprintAspectRatio, 
                                               lineWidth = self.config.narrowedLineWidth, 
                                               display = display, 
                                               debug = debug, 
                                               )

        if goodMask is None:
            goodMask = maskTmp.getPlaneBitMask("DETECTED")
        if badMask is None:
            badMaskList = ['NO_DATA', 'EDGE', 'BAD', 'SAT', 'INTRP', 'CR']
            badMask = maskTmp.getPlaneBitMask(badMaskList)

        self.log.info('goodMaskBit = %d, badMaskBit = %d' % (goodMask, badMask))


        # calculate skysigma and threshold for pixels used in Hough
        if minImageValue < 0 or maxImageValue < 0:
            maskVal = 0
            for maskBit in badMaskList:
                maskVal |= afwImage.MaskU.getPlaneBitMask(maskBit)
            nClip = 3.0; nIter = 3
            sctrl = afwMath.StatisticsControl(nClip, nIter, maskVal)
            stat = afwMath.makeStatistics(mi, afwMath.MEANCLIP | afwMath.STDEVCLIP | afwMath.MAX, sctrl)
            bkg = stat.getValue(afwMath.MEANCLIP)
            bkgSigma = stat.getValue(afwMath.STDEVCLIP)

            if minImageValue < 0:
                if lowSigma > 0:
                    minImageValue = bkg + lowSigma * bkgSigma
            if maxImageValue < 0:
                if highSigma > 0:
                    maxImageValue = bkg + highSigma * bkgSigma
                else:
                    maxImageValue = stat.getValue(afwMath.MAX)

        else: # use the given minImageValue and maxImageValue
            pass
    
        #
        # perform Hough transformation
        #
        maskArr = maskTmp.getArray()
        maskArr = numpy.array(maskArr, dtype=numpy.int32)
        maskHough = afwImage.ImageI(maskArr)

        if display:
            ds9.mtv(maskHough, frame = 1)
        if self.config.doWriteFile:
            print self.config.outputFile
            maskHough.writeFits(self.config.outputFile + '_mask4hough.fits')

        self.log.info('nAngle: %s nRho: %s' % (nAngle, nRho))
        hough = qa.Hough(maskHough, nAngle, nRho)

        if minImageValue is not None:
            self.log.info('image count threshold applied to Hough transform: (%s, %s) (count)' % (minImageValue, maxImageValue))
            im = mi.getImage()
            himage, ang, dis = hough.calcHough(im, minImageValue, maxImageValue, goodMask, badMask)
        else:
            self.log.info('image count threshold not applied to Hough transform')
            himage, ang, dis = hough.calcHough(goodMask, badMask)

        self.log.info('Npix effective: %s' % hough.getNpixEffective())

        self.himage = himage

        # This is the highest peak returned by in Hough class 
        # But we need to detect all possible peaks corresponding to trails, so do not simply use this
        a = -1.0 * (numpy.cos(ang)/numpy.sin(ang))
        b = dis / numpy.sin(ang)
        # print ang, dis
        # print "y = %g * x + %g" % (a, b)

        return self.himage, fpList
    

    def getHoughPeaks(self, mask, himage, minHough, allowedPeakRatio = None,
                      minPeakArea = 10,  minPeakCounts = 10000, rMargin1 = 10.0, rMargin2 = 60.0, 
                      minHoughPerFootprint = None, 
                      debug = False, 
                      display = False, 
                      ):
        """
        select all possible peaks (x, y) in resultant Hough-transformed plane
        """

        # this is a piece of experiments of scaling conditions 
        # based on 2k4k CORR experience to other-size data
        if False:
            maskArr = mask.getArray()
            maskEff = maskArr[ 
                numpy.logical_and((maskArr & mask.getPlaneBitMask("NO_DATA") == 0), 
                                  (maskArr & mask.getPlaneBitMask("EDGE") == 0)) 
                ]

            scale = float(len(maskEff)) / (2048.*4176.)
            if scale < 1: # if the effective area is smaller than 2k4k, we will use 1 as the lowest scale
                scale = 1.0
            minHough *= scale # scaled by effective image area
            self.log.info('effective area: %d (pix) / (2048*4176) = %8.4f : scale_used = %8.4f' % (len(maskEff), float(len(maskEff)) / (2048.*4176.), scale))

        # for experiments of scaling conditions by number of footprints
        nfp = float(len(self.fpList))

        vHough = himage.getArray()
        nyHough, nxHough = vHough.shape
        self.log.info('Hough nx: %d ny: %d' % (nxHough, nyHough))

        # convert hough to 1d array 
        vHough1d = numpy.reshape(vHough, nxHough * nyHough )
        x1dInd = numpy.array( [range(nxHough)] * nyHough, dtype=numpy.int32 )
        y1dInd = numpy.array( [ [y] * nxHough for y in range(nyHough) ], dtype=numpy.int32 )
        x1dInd = numpy.reshape(x1dInd, nxHough * nyHough )
        y1dInd = numpy.reshape(y1dInd, nxHough * nyHough )

        # pick out only high hough pixels
        nCand = len(vHough1d)
        self.log.info('trail candidates before filtering: %s' % nCand)
        if minHoughPerFootprint > 0:
            indHigh = numpy.logical_and(vHough1d > int(nfp*minHoughPerFootprint), y1dInd != 500) # clipping 500 is workaround
        else:
            indHigh = numpy.logical_and(vHough1d > minHough, y1dInd != 500) # clipping 500 is workaround
        vHough1dHigh = vHough1d[indHigh]
        x1dIndHigh = x1dInd[indHigh]
        y1dIndHigh = y1dInd[indHigh]

        # grouping high hough-pixel peaks, 
        #   assigning one pixel for each hough peak group by taking the highest hough value in a peak group
        xCandList = numpy.array([])
        yCandList = numpy.array([])

        if len(vHough1dHigh) == 0:
            self.log.info('trail candidates after filtering: %s' % 0)
            self.log.info('No trail candidate found.')
            # display the resultant Hough peaks on ds9
            if display:
                ds9.mtv(himage, frame = 3)
            if self.config.doWriteFile:
                himage.writeFits(self.config.outputFile + '_hough.fits')

            return xCandList, yCandList 

        ind1dHighSorted = numpy.argsort(vHough1dHigh)
        #   revert to make it descending
        ind1dHighSorted = ind1dHighSorted[-1::-1]
        vHough1dHighSorted = vHough1dHigh[ind1dHighSorted]
        x1dIndHighSorted = x1dIndHigh[ind1dHighSorted]
        y1dIndHighSorted = y1dIndHigh[ind1dHighSorted]

        # further limit pixels by the hough's peak ratio
        maxPeak = vHough1dHighSorted[0]
        if allowedPeakRatio > 0.0:
            minPeak = maxPeak / allowedPeakRatio
            self.log.info('allowedPeakRatio = %f' %  allowedPeakRatio)
        else:
            minPeak = 0
            self.log.info('allowedPeakRatio unset.')
        self.log.info('maxPeak = %f' % maxPeak)
        self.log.info('minPeak = %f' % minPeak)
        vHough1dFiltered = vHough1dHighSorted
        x1dIndFiltered = x1dIndHighSorted
        y1dIndFiltered = y1dIndHighSorted

        #   start grouping hough peak groups, with descending-sorted vHoughArr
        vHoughArr = numpy.copy(vHough1dFiltered)
        xArr = numpy.copy(x1dIndFiltered)
        yArr = numpy.copy(y1dIndFiltered)

        while len(vHoughArr) > 0:

            if debug:
                self.log.debug('+' * 30)
                for x, y, v in zip(xArr, yArr, vHoughArr):
                    self.log.debug('%s %s %s %s' % (x, y, v, v/nfp))
                self.log.debug('+' * 30)

            peakThisKnot = vHoughArr[0] # highest peak in this group (knot)
            x0 = xArr[0]
            y0 = yArr[0]
            
            # count up hough points which belongs to this peak group, and
            # carry over the other groups to the next loop
            nPixelsThisKnot = 0
            countsThisKnot = 0
            indOthers = numpy.array([], dtype = int)
            for i, (x, y, v) in enumerate(zip(xArr, yArr, vHoughArr)):
                dist = numpy.hypot(x - x0, y - y0)
                if dist < rMargin1: # counted in this knot group
                    nPixelsThisKnot += 1
                    countsThisKnot += v
                if dist > rMargin2: # carried to other knot groups
                    indOthers = numpy.append(indOthers, i)

            # decide if this group is a significant trail
            #if (peakThisKnot >= minPeak) and (nPixelsThisKnot >= minPeakArea) and (countsThisKnot >= minPeakCounts):
            #if (nPixelsThisKnot >= minPeakArea): # minPeak is already screened by minhough_per_footprint
            if (nPixelsThisKnot >= minPeakArea) and (countsThisKnot >= minPeakCounts):
                xCandList = numpy.append(xCandList, x0)
                yCandList = numpy.append(yCandList, y0)

                self.log.info('thisKnot(%s, %s)(peak, npixel, counts): (%s, %s, %s) nfp=%d (%s, %s, %s) -- selected' % (x0, y0, peakThisKnot, nPixelsThisKnot, countsThisKnot, nfp, peakThisKnot/nfp, nPixelsThisKnot/nfp, countsThisKnot/nfp))
            else:
                if debug:
###                    print 'thisTestedKnot(%s, %s)(peak, npixel, counts): (%s, %s, %s) -- rejected' % (x0, y0, peakThisKnot, nPixelsThisKnot, countsThisKnot)
                    self.log.info('thisTestedKnot(%s, %s)(peak, npixel, counts): (%s, %s, %s) nfp=%d (%s, %s, %s) -- rejected' % (x0, y0, peakThisKnot, nPixelsThisKnot, countsThisKnot, nfp, peakThisKnot/nfp, nPixelsThisKnot/nfp, countsThisKnot/nfp))

            vHoughArr = vHoughArr[indOthers]
            xArr = xArr[indOthers]
            yArr = yArr[indOthers]

        nCandHigh = len(xCandList)
        self.log.info('trail candidates after filtering: %s' % nCandHigh)


        # display the resultant Hough peaks on ds9
        if display:
            ds9.mtv(himage, frame = 3)
            for (t, r) in zip(xCandList, yCandList):
                ds9.dot('o', t, r, frame = 3, size = rMargin2, ctype = 'cyan')
        if self.config.doWriteFile:
            himage.writeFits(self.config.outputFile + '_hough.fits')

        return xCandList, yCandList


    def convertHoughToTrails(self, mask, xHoughPeakList, yHoughPeakList, nRho, nAngle):
        """
        convert input Hough peaks into line parameters (a:slope, b:intersect)
        """
    
        nxImage, nyImage = mask.getDimensions()

        anglePeakList = numpy.pi * xHoughPeakList / nAngle 
        rhoPeakList = (yHoughPeakList - float(nRho)) * int(numpy.hypot(nxImage-1, nyImage-1) + 1) / float(nRho)

        anglePeakList[ anglePeakList < 0.0001 ] = 0.0001 # workaround for avoid singlarity of (angle=0 a=inf); 0.0001rad ~20arcsec = 0.4px offset between 4k CCD edges
        
        aList = -1.0 * (numpy.cos(anglePeakList)/numpy.sin(anglePeakList))
        bList = rhoPeakList / numpy.sin(anglePeakList)

        return aList, bList


    def getTrailFootprints(self, mask, fpList, aList, bList, trailMargin, minTrailArea, debug = False):
        """ select footprints of trail candidates from the input footprint list
        """

# check if this footprint overlaps with the trails candidates in each y 

        xmin, ymin = mask.getBBox(afwImage.PARENT).getMin()

        fpTrailList = []
        for fp in fpList:

            # footprint should be larger than the given minArea to be a trail
            if debug:
                self.log.info('trailArea: %s' % fp.getArea())
            if fp.getArea() < minTrailArea:
                continue

            # check if this fp overlaps with trail candidates
            for s in fp.getSpans():
                x0, x1 = s.getX0() -xmin, s.getX1() -xmin
                y = s.getY() - ymin

                # define trail regions with margin for recognizing this footprint as a candidate
                xTrailList = (y - bList) / aList
                minXTrailList, maxXTrailList  = xTrailList - trailMargin, xTrailList + trailMargin

                if ( numpy.logical_and(minXTrailList < x1, maxXTrailList > x0).any() ):
                    fpTrailList.append(fp)
                    break

        return fpTrailList


    def maskTrailCandidates(self, exp, aList, bList, detectedBit, trailBit, 
                            trailMargin = 30.0, minTrailArea = 300, display = False, debug = False, 
                            ):
        """ select footprints of possible trails and set new mask bit to them 
        """

        mask = exp.getMaskedImage().getMask()
        thresh = afwDetection.Threshold(detectedBit, afwDetection.Threshold.BITMASK)
        fpList = afwDetection.FootprintSet(mask, thresh).getFootprints()

        xmin, ymin = mask.getBBox(afwImage.PARENT).getMin()

        fpTrailList = self.getTrailFootprints(mask, fpList, aList, bList, trailMargin, minTrailArea, debug = debug)
    
        for fp in fpTrailList:
            for s in fp.getSpans():
                x0, x1 = s.getX0() - xmin, s.getX1() - xmin # we may handle patch so need to care offset
                y = s.getY() - ymin
                for x in range(x0, x1):
                    mask.set(x, y, mask.get(x, y) | trailBit)


        # display the resultant mask on ds9
        if display:
            ds9.setMaskPlaneColor('TRAIL', 'cyan')
            ds9.mtv(mask.convertU(), frame = 4)

        return fpTrailList


class HscTrailMaskConfig(HscTrailConfig):
    pass

class HscTrailMaskTask(HscTrailTask):
    """
    A task to detect and mask satellite trails called from CmdLineTask or BatchPoolTask.
    """

    _DefaultName = 'mask_trail'
    ConfigClass = HscTrailMaskConfig

    def __init__(self, *args, **kwargs):
        super(HscTrailMaskTask, self).__init__(*args, **kwargs)


    def run(self, dataRef, exp, display = False, debug = False):
        """
        detect and mask candidates of satellite trails.
        """

        super(HscTrailMaskTask, self).run(exp, display, debug)

