import sys, os
import math
import lsst.daf.persistence as dafPersist
import lsst.obs.hsc as obsHsc
import lsst.pex.config as pexConfig
import lsst.pipe.base as pipeBase
import lsst.afw.image as afwImage
import lsst.afw.geom as afwGeom
import lsst.afw.cameraGeom as cameraGeom
import lsst.afw.detection as afwDetection
import lsst.afw.coord as afwCoord

from lsst.meas.astrom import Astrometry
from lsst.meas.astrom.config import MeasAstromConfig, AstrometryNetDataConfig


import numpy

### input parameters
# ra_t_d : right ascension of the telescope pointing in degree
# de_t_d : declination of the telescope pointing in degree
# ra_s_d : right ascension of the guide star in degree
# de_s_d : declination of the guide star in degree
# lst_d  : local sidereal time in degree
# mjd    : modified julian date


class HscMustacheGhostConfig(pexConfig.Config):
    """
    Config class for predicting 'mustache' ghosts
    """

    offsetX = pexConfig.Field(dtype = float, default = 12.78, 
                              doc = "Offset of mother-plate coord x to obs_subaru FOV center")
    offsetY = pexConfig.Field(dtype = float, default = 57.74, 
                              doc = "Offset of mother-plate coord y to obs_subaru FOV center")
    adjusting_factor = pexConfig.DictField(keytype = str, itemtype = float, 
                                           default = {
            'g':1.002, 'r':1.0015, 'i':1.001, 'z':0.9995, 'y':1.000, 'N921':1.0005
            }, 
                                           doc = "Filter dependency of ghost location")
    sampling_step = pexConfig.Field(dtype = float, default = 0.25, 
                                    doc = "Sampling step to cover each ghost arc")
    minRadius = pexConfig.Field(dtype = float, default = 0.8, 
                                doc = "Minimum radius to extract reference sources as ghosts origin")
    maxRadius = pexConfig.Field(dtype = float, default = 1.5, 
                                doc = "Maximum radius to extract reference sources as ghosts origin")
    filterNameRef = pexConfig.DictField(keytype = str, itemtype = str, 
                                        default = {
            'g':'g', 'r':'r', 'i':'i', 'i2':'i', 'r2':'r', 'z':'z', 'y':'y', 
            'N656':'r', 'N921':'z', 'N816':'i', 'N387':'g', 'N101':'y', 
                                                   }, 
                                        doc = "Mapping of filter names for reference and ghost prediction")
    maglim = pexConfig.DictField(keytype = str, itemtype = float, default = {
            'g':13.0, 'r':13.5, 'i':13.5, 'z':13.0, 'y':13.0, 'N921':12.0, 'N656': 13.5, 'N816':13.5,  'N387': 13.0, 
            },  
                                 doc = "Limiting magnitude to calculate ghost prediction")
    minGhostArea = pexConfig.Field(dtype = float, default = 10000.0, 
                                    doc = "Minimum area allowed for each ghost footprint")
    marginGhostPosition = pexConfig.Field(dtype = float, default = 30.0, 
                                    doc = "Consider any footprints located within this number x 3 to mask ghosts")


    def setDefaults(self):
        super(HscMustacheGhostConfig, self).setDefaults()


class HscMustacheGhostTask(pipeBase.Task):
    """
    A task to derive a list of pixels where the mustache ghosts may be located due to bright stars.
    Credit:
      The codes for calculating the ghost locations are provided by 
      Dr. Satoshi Kawanomoto, Subaru Telescope/HSC project, NAOJ
    """

    _DefaultName = 'ghost_mustache'
    ConfigClass = HscMustacheGhostConfig

    def __init__(self, *args, **kwargs):
        super(HscMustacheGhostTask, self).__init__(*args, **kwargs)

    ## fixed parameters
    _pixscaleMm = 0.015 # Pixel scale in mm
    _ltt_d = 19.82556   # Dome latitude in degree
    _lng_d = -155.47611 # Dome longitude in degree
    _mjd_J2000 = 51544.5 # MJD at J2000.0 (2000/01/01.5)
    # T=273.15[K], P=600[hPa], Pw=1.5[hPa], lambda=0.55[um]
    _air_idx = 1.0 + 1.7347e-04 # Refraction index of air 
    _air_sh = 0.00130  # scale height of air

    def mjd2jc2000(self, mjd):
        """convert mjd to Julian century (J2000.0 origin)"""
        jc2000 = (mjd - self._mjd_J2000) / 36525.0
        return jc2000

    def precessionMatrix(self, jc2000):
        """create precession matrix at the given time in Julian century"""
        zeta_A  = numpy.deg2rad((2306.2181*jc2000 + 0.30188*jc2000**2.0 + 0.017998*jc2000**3.0)/3600.0)
        z_A     = numpy.deg2rad((2306.2181*jc2000 + 1.09468*jc2000**2.0 + 0.018203*jc2000**3.0)/3600.0)
        theta_A = numpy.deg2rad((2004.3109*jc2000 - 0.42665*jc2000**2.0 - 0.041833*jc2000**3.0)/3600.0)
        precMat = numpy.matrix([[+numpy.cos(zeta_A)*numpy.cos(theta_A)*numpy.cos(z_A)-numpy.sin(zeta_A)*numpy.sin(z_A),\
                              -numpy.sin(zeta_A)*numpy.cos(theta_A)*numpy.cos(z_A)-numpy.cos(zeta_A)*numpy.sin(z_A),\
                              -numpy.sin(theta_A)*numpy.cos(z_A)],
                             [+numpy.cos(zeta_A)*numpy.cos(theta_A)*numpy.sin(z_A)+numpy.sin(zeta_A)*numpy.cos(z_A),\
                              -numpy.sin(zeta_A)*numpy.cos(theta_A)*numpy.sin(z_A)+numpy.cos(zeta_A)*numpy.cos(z_A),\
                              -numpy.sin(theta_A)*numpy.sin(z_A)],
                             [+numpy.cos(zeta_A)*numpy.sin(theta_A)                                       ,\
                              -numpy.sin(zeta_A)*numpy.sin(theta_A)                                       ,\
                              +numpy.cos(theta_A)]])
        return precMat

    def mjd2gmst(self, mjd):
        """convert mjd to GMST(Greenwich mean sidereal time)"""
        mjd_f = mjd%1
        jc2000 = self.mjd2jc2000(mjd)
        gmst_s = (6.0*3600.0+41.0*60.0+50.54841)\
                  + 8640184.812866*jc2000 + 0.093104*jc2000**2.0 - 0.0000062*jc2000**3.0\
                  + mjd_f*86400.0
        gmst_d = (gmst_s%86400)/240.0
        return gmst_d

    def gmst2lmst(self, gmst_d):
        """convert GMST to LMST(local mean sidereal time)"""
        lmst_d = (gmst_d + self._lng_d)%360
        return lmst_d

    def sph2vec(self, ra_d, de_d):
        """convert spherical coordinate to the Cartesian coordinates (vector)"""
        ra_r = numpy.deg2rad(ra_d)
        de_r = numpy.deg2rad(de_d)
        vec  = numpy.array([[numpy.cos(ra_r)*numpy.cos(de_r)],
                         [numpy.sin(ra_r)*numpy.cos(de_r)],
                         [             numpy.sin(de_r)]])
        return vec

    def vec2sph(self, vec):
        """convert the Cartesian coordinates vector to shperical coordinates"""
        ra_r = numpy.arctan2(vec[1,0],vec[0,0])
        de_r = numpy.arcsin(vec[2,0])
        ra_d = numpy.rad2deg(ra_r)
        de_d = numpy.rad2deg(de_r)
        return ra_d,de_d

    def ra2ha(self, ra_d, lst_d):
        """convert right ascension to hour angle at given LST"""
        ha_d = (lst_d - ra_d)%360
        return ha_d

    def eq2hz(self, ha_d, de_d):
        """convert equatorial coordinates to the horizontal coordinates"""
        ltt_r = numpy.deg2rad(self._ltt_d)
        ha_r  = numpy.deg2rad(ha_d)
        de_r  = numpy.deg2rad(de_d)
        zd_r  = numpy.arccos(+numpy.sin(ltt_r)*numpy.sin(de_r)+numpy.cos(ltt_r)*numpy.cos(de_r)*numpy.cos(ha_r))
        az_r  = numpy.arctan2(+numpy.cos(de_r)*numpy.sin(ha_r),
                           -numpy.cos(ltt_r)*numpy.sin(de_r)+numpy.sin(ltt_r)*numpy.cos(de_r)*numpy.cos(ha_r))
        zd_d  = numpy.rad2deg(zd_r)
        az_d  = numpy.rad2deg(az_r)
        al_d  = 90.0 - zd_d
        return al_d, az_d

    def air_idx(self):
        """return the air refraction index"""
        return self._air_idx

    def atm_ref(self, al_d):
        """return the atmospheric refraction at given altitude"""
        if al_d > 20.0:
            zd_r = numpy.deg2rad(90.0-al_d)
        else:
            zd_r = numpy.deg2rad(70.0)
        r0 = self.air_idx()-1.0
        sh = self._air_sh
        R0 = (1.0-sh)*r0-sh*r0**2/2.0+sh**2*r0*2.0
        R1 = r0**2/2.0+r0**3/6.0-sh*r0-sh*r0**2*11.0/4.0+sh**2*r0*5.0
        R2 = r0**3-sh*r0**2*9.0/4.0+sh**2*r0*3.0
        R = R0*numpy.tan(zd_r)+R1*(numpy.tan(zd_r))**3+R2*(numpy.tan(zd_r))**5
        return numpy.rad2deg(R)

    def mal2aal(self, mal_d):
        """convert mean altitude to apparent altitude"""
        aal_d = mal_d + self.atm_ref(mal_d)
        return aal_d

    def pos2adt(self, al_t_d, al_s_d, delta_az_d):
        """convert altitudes of telescope and star and relative azimuth to angular distance and position angle"""
        zd_t_r = numpy.deg2rad(90.0-al_t_d)
        zd_s_r = numpy.deg2rad(90.0-al_s_d)
        daz_r  = numpy.deg2rad(delta_az_d)

        ad_r   = numpy.arccos(numpy.cos(zd_t_r)*numpy.cos(zd_s_r)+numpy.sin(zd_t_r)*numpy.sin(zd_s_r)*numpy.cos(daz_r))

        a = zd_s_r
        b = zd_t_r
        c = ad_r
        s = 0.5*(a+b+c)

        if ad_r>0.0:
            pa_r = 2.0*numpy.arcsin(numpy.sqrt(numpy.sin(s-b)*numpy.sin(s-c)/(numpy.sin(b)*numpy.sin(c))))
            if daz_r<0.0:
                pa_r = 2.0*numpy.pi-pa_r
        else:
            pa_r = 0.0

        ad_d = numpy.rad2deg(ad_r)
        pa_d = numpy.rad2deg(pa_r)

        return ad_d,pa_d

    def ghostSourcePosWithLST(self, ra_t_d, de_t_d, ra_s_d, de_s_d, inr_d, mjd, lst):
        """return the position on the cold plate of given guide star and telescope direction"""
        jc2000 = self.mjd2jc2000(mjd)
        pm     = self.precessionMatrix(jc2000)

        vt      = self.sph2vec(ra_t_d, de_t_d)
        vt_mean = numpy.dot(pm,vt)
        vs      = self.sph2vec(ra_s_d, de_s_d)
        vs_mean = numpy.dot(pm,vs)

        (mean_ra_t_d, mean_de_t_d) = self.vec2sph(vt_mean)
        (mean_ra_s_d, mean_de_s_d) = self.vec2sph(vs_mean)

        lmst_d = lst

        mean_ha_t_d = self.ra2ha(mean_ra_t_d,lmst_d)
        mean_ha_s_d = self.ra2ha(mean_ra_s_d,lmst_d)

        (mean_al_t_d, mean_az_t_d) =  self.eq2hz(mean_ha_t_d, mean_de_t_d)
        (mean_al_s_d, mean_az_s_d) =  self.eq2hz(mean_ha_s_d, mean_de_s_d)

        apparent_al_t_d = self.mal2aal(mean_al_t_d)
        apparent_al_s_d = self.mal2aal(mean_al_s_d)

        delta_az_d = mean_az_t_d - mean_az_s_d

        (ang_dist_d, p_ang_d) = self.pos2adt(apparent_al_t_d, apparent_al_s_d, delta_az_d)

        return ang_dist_d, p_ang_d

    def arcCR(self, band, ad_d):

        # HSC-g
        #  4467.9759580157 -10022.1277915657  5156.3556611527
        # -4375.9715493905   9717.6993803237 -5203.2424778605
        # HSC-r2
        #  4354.2575893763  -9762.9766653532  5011.1243975288
        # -4263.2230039887   9457.4516871365 -5057.2762518308
        # HSC-i2
        #  4243.1837126908  -9517.3512135240  4876.3237965836
        # -4154.5095389959   9215.4140278639 -4924.2923196315
        # HSC-z
        #  3966.2315539829  -8920.7199277389  4555.9567436784
        # -3878.4529585585   8619.7241050446 -4604.3603386145
        # HSC-Y
        #  3723.5919271089  -8403.7641770854  4281.0159171765
        # -3637.0352789862   8104.8245491120 -4330.4583627761
        parameters = {
            'g': [ -4467.9759580157,  10022.1277915657,  -5156.3556611527, -4375.9715493905,   9717.6993803237, -5203.2424778605],
            'r': [ -4354.2575893763,   9762.9766653532,  -5011.1243975288, -4263.2230039887,   9457.4516871365, -5057.2762518308],
            'i': [ -4243.1837126908,   9517.3512135240,  -4876.3237965836, -4154.5095389959,   9215.4140278639, -4924.2923196315],
            'z': [ -3966.2315539829,   8920.7199277389,  -4555.9567436784, -3878.4529585585,   8619.7241050446, -4604.3603386145],
            'y': [ -3723.5919271089,   8403.7641770854,  -4281.0159171765, -3637.0352789862,   8104.8245491120, -4330.4583627761],
            }
        C = parameters[band][0]+parameters[band][1]*ad_d+parameters[band][2]*ad_d**2
        R = parameters[band][3]+parameters[band][4]*ad_d+parameters[band][5]*ad_d**2

        return C,R

    def outerEdgeAngle(self, ad_d):
        if   ad_d < 0.8831:
            oea = 0.0
        elif ad_d < 0.9368:
            oea = 46.2985*numpy.sqrt(ad_d-0.8831)
        else:
            oea = -2.6788646406597 + 14.311535418553*ad_d

        return oea

    def innerEdgeAngle(self, ad_d):
        if ad_d < 0.96464:
            iea = -243.86606061927*(ad_d-0.93238966579931)**2.0+2.4636017943593
        else:
            iea = 66.2061*numpy.sqrt(ad_d-0.963531)

        return iea


    def getGhostsMm(self, band, mjd, ra_t_d, de_t_d, inr_d, ra_s_d, de_s_d):
        """ return possible ghost coordinates in mm on the mother plate coordinates
        """

        lst = self.gmst2lmst(self.mjd2gmst(mjd))

        (ad,pa) = self.ghostSourcePosWithLST(ra_t_d, de_t_d, ra_s_d, de_s_d, inr_d, mjd, lst)

        (center, radius) = self.arcCR(band, ad)

        out = self.outerEdgeAngle(ad)
        inn = self.innerEdgeAngle(ad)

        points = numpy.arange(inn, out, self.config.sampling_step)
        n = points.shape[0]

        x = center-radius*numpy.cos(numpy.deg2rad(points))
        y =        radius*numpy.sin(numpy.deg2rad(points))

        rot = numpy.deg2rad(90.0+pa+inr_d)

        arc1_x = x*numpy.cos(rot)-y*numpy.sin(rot)
        arc1_y = y*numpy.cos(rot)+x*numpy.sin(rot)

        arc2_x = x*numpy.cos(rot)+y*numpy.sin(rot)
        arc2_y =-y*numpy.cos(rot)+x*numpy.sin(rot)

        ghostList = []
        for idx in range(n):
            ghostList.append([arc1_x[idx],arc1_y[idx]])
        for idx in range(n):
            ghostList.append([arc2_x[idx],arc2_y[idx]])

        return ghostList

    def getGhostsCcd(self, butler, exp, band, ccdId, mjd, ra_t_d, de_t_d, inr_d, ra_s_d, de_s_d, display = False, outputReg = True, debug = False):
        """ get CCD local coordinates of possible ghosts for a bright star (ra_s_d, de_s_d)
        """

        #print 'getGhostsCcd: band = %s' % band

        ghostListMm = self.getGhostsMm(band, mjd, ra_t_d, de_t_d, inr_d, ra_s_d, de_s_d)

        # small adjustment for band compensating residuals from the ghost model
        tmpList = ghostListMm
        ghostListMm = []
        for x, y in tmpList:
            x *= self.config.adjusting_factor[band]
            y *= self.config.adjusting_factor[band]
            ghostListMm.append([x, y])

        # convert mother-plate coords (mm) to global coords (px)
        xList_global = []
        yList_global = []
        for x_mp, y_mp in ghostListMm:
            x_global = x_mp * numpy.cos(-numpy.pi/2.) + y_mp * numpy.sin(-numpy.pi/2.)
            y_global = - x_mp * numpy.sin(-numpy.pi/2.) + y_mp * numpy.cos(-numpy.pi/2.)

            x_global = x_global / self._pixscaleMm + self.config.offsetX
            y_global = y_global / self._pixscaleMm + self.config.offsetY

            xList_global.append(x_global)
            yList_global.append(y_global)

        camera = butler.get("camera")
        raft = camera[0]
        raft = cameraGeom.cast_Raft(raft)
        ccd = raft.findDetector(cameraGeom.Id(ccdId))

        x0Ccd, y0Ccd = ccd.getPositionFromPixel(afwGeom.PointD(0.0, 0.0)).getMm()
        xHalfWidth, yHalfHeight = ccd.getCenterPixel()

        ccd.setTrimmed(True)

        # global position of this CCD
        #Point2D(1023.5, 2087.5)
        #xCenCcd, yCenCcd = ccd.getPositionFromPixel(afwGeom.PointD(0.0, 0.0)).getMm()
        #x0Ccd, y0Ccd = ccd.getPositionFromPixel(afwGeom.PointD(0.0, 0.0)).getMm()
        rotCcd = ccd.getOrientation().getYaw().asRadians() # radian

        # mapping ghost positions (mm) on to global coords
        ghostListXy_local = []
        for x_global, y_global in zip(xList_global, yList_global):

            # CCD local
            if False:
                xCcd_local_tmp = x_global - x0Ccd
                yCcd_local_tmp = y_global - y0Ccd
                #(x_local) =  (xCcd_local_tmp) * (numpy.cos(rotCcd) -sin(rotCcd))
                #(y_local)    (yCcd_local_tmp)   (sin(rotCcd)  numpy.cos(rotCcd))
                xCcd_local =  xCcd_local_tmp * numpy.cos(rotCcd) + yCcd_local_tmp * numpy.sin(rotCcd)
                yCcd_local = -xCcd_local_tmp * numpy.sin(rotCcd) + yCcd_local_tmp * numpy.cos(rotCcd)
            else:
                xCcd_local_tmp = x_global - x0Ccd
                yCcd_local_tmp = y_global - y0Ccd
                #(x_local) =  (xCcd_local_tmp) * (numpy.cos(rotCcd) -sin(rotCcd))
                #(y_local)    (yCcd_local_tmp)   (sin(rotCcd)  numpy.cos(rotCcd))
                xCcd_local =  xCcd_local_tmp * numpy.cos(rotCcd) + yCcd_local_tmp * numpy.sin(rotCcd) #+ xHalfWidth
                yCcd_local = -xCcd_local_tmp * numpy.sin(rotCcd) + yCcd_local_tmp * numpy.cos(rotCcd) #+ yHalfHeight

            # ensure this ghost prediction circle overlaps with at least a part of the CCD pixels
            xmax_ccd = exp.getWidth() - 1.; xmin_ccd = 0.0
            ymax_ccd = exp.getHeight() -1.; ymin_ccd = 0.0

            if debug:
                print "(%f, %f) - (%f, %f)" % (xmax_ccd, ymax_ccd, xCcd_local, yCcd_local)

            if (xCcd_local > xmin_ccd) and (xCcd_local < xmax_ccd) and \
                    (yCcd_local > ymin_ccd) and (yCcd_local < ymax_ccd):

                ghostListXy_local.append([xCcd_local, yCcd_local])

        return ghostListXy_local


    def run(self, butler, band, ccdId, mjd, ra_t_d, de_t_d, inr_d, ra_s_d, de_s_d, display = False, outputReg = True):
        self.getGhostsCcd(butler, exp, band, ccdId, mjd, ra_t_d, de_t_d, inr_d, ra_s_d, de_s_d, display = display, outputReg = outReg)


class HscCausticGhostConfig(pexConfig.Config):
    """
    Config class for predicting 'causitc' ghosts
    """

    offsetX = pexConfig.Field(dtype = float, default = 12.78, 
                              doc = "Offset of mother-plate coord x to obs_subaru FOV center")
    offsetY = pexConfig.Field(dtype = float, default = 57.74, 
                              doc = "Offset of mother-plate coord y to obs_subaru FOV center")
    adjusting_factor = pexConfig.DictField(keytype = str, itemtype = float, 
                                           default = {
            'g':1.002, 'r':1.0015, 'i':1.001, 'z':0.9995, 'y':1.000, 'N921':1.0005
            }, 
                                           doc = "Filter dependency of ghost location")
    sampling_step = pexConfig.Field(dtype = float, default = 0.25, 
                                    doc = "Sampling step to cover each ghost arc")
    minRadius = pexConfig.Field(dtype = float, default = 0.0, 
                                doc = "[deg] Minimum radius to extract reference sources as ghosts origin")
    maxRadius = pexConfig.Field(dtype = float, default = 0.55, 
                                doc = "[deg] Maximum radius to extract reference sources as ghosts origin")
    filterNameRef = pexConfig.DictField(keytype = str, itemtype = str, 
                                        default = {
            'g':'g', 'r':'r', 'i':'i', 'i2':'i', 'r2':'r', 'z':'z', 'y':'y', 
            'N656':'r', 'N921':'z', 'N816':'i', 'N387':'g', 'N101':'y', 
                                                   }, 
                                        doc = "Mapping of filter names for reference and ghost prediction")
    magFaint = pexConfig.DictField(keytype = str, itemtype = float, default = {
            'g':12.0, 'r':11.5, 'i':11.0, 'z':11.0, 'y':11.0, 'N921':11.0, 'N656': 11.0, 'N816':11.0,  'N387': 12.0, 
            }, 
                                 doc = "Faintest magnitude to calculate ghost prediction")
    magBright = pexConfig.DictField(keytype = str, itemtype = float, default = {
            'g':-2., 'r':-2., 'i':-2., 'z':-2., 'y':-2., 'N921':-2., 'N656': -2., 'N816':-2., 'N387': -2., 
            }, 
                                 doc = "Brightest magnitude to calculate ghost prediction")
    minGhostArea = pexConfig.Field(dtype = float, default = 3000.0, 
                                    doc = "Minimum area allowed for each ghost footprint")
    marginGhostPosition = pexConfig.Field(dtype = float, default = 30.0, 
                                    doc = "Consider any footprints located within this number x 3 to mask ghosts")
    def setDefaults(self):
        super(HscCausticGhostConfig, self).setDefaults()


class HscCausticGhostTask(pipeBase.Task):
    """
    A task to derive a list of pixels where the 'caustic' ghosts may be located, 
    which are caused by bright stars inside the FOV.

    Credit:
      The model for calculating locations of the caustic ghosts are provided by 
      Dr. Yutaka Komiyama, Subaru Telescope/HSC project, NAOJ
    """

    _DefaultName = 'ghost_caustic'
    ConfigClass = HscCausticGhostConfig

    def __init__(self, *args, **kwargs):
        super(HscCausticGhostTask, self).__init__(*args, **kwargs)

    ## fixed parameters
    _pixscaleMm = 0.015 # Pixel scale in mm

    CP_CR_dict = {
        'g': [
            [ -8.66581086e+04,   1.60789252e+05,  -1.19834803e+05,   4.35455723e+04,
               -8.20984903e+03,   5.94514717e+02,  -4.56480396e+02,   3.86129097e-04], 
            [ -3.66203193e+07,   8.71449290e+07,  -7.68658356e+07,   3.21678893e+07,
               -6.59305008e+06,   6.25378363e+05,  -1.14163334e+04,   2.80257250e+03]
            ],
        'r': [
            [  8.14947896e+03,  -1.24163720e+04,  5.41143508e+03,  -1.02024286e+03,
               -2.39667209e+02,  -1.00805716e+02,  -4.16757016e+02,  -1.48799186e-03], 
            [ -2.34342573e+07,   3.47643396e+07,  -1.98155551e+07,   5.75428423e+06,
               -9.54233470e+05,   9.14792638e+04,   1.45742667e+03,   6.96530721e+03]
            ],
        'r2': [
            [  1.05167578e+04, -1.55470840e+04,   7.03878935e+03,  -1.44358813e+03,
               -1.75333116e+02,  -1.08162329e+02,  -4.13994264e+02,   1.10217740e-03],
            [ -2.46004825e+07,   3.55382290e+07,  -1.97396910e+07,   5.64254906e+06,
               -9.45421731e+05,   9.57272709e+04,  -1.34007061e+01,   7.84632263e+03]
            ],
        'i': [
            [  7.09135701e+03,  -1.08899009e+04,   4.53311864e+03,  -7.45070274e+02,
               -2.86196488e+02,  -9.58537252e+01,  -4.15997892e+02,  -2.10038501e-03],
            [ -2.65938293e+07,   3.98815563e+07,  -2.30650423e+07,   6.77205238e+06,
               -1.11878753e+06,   1.03127466e+05,   1.44010771e+03,   6.83384243e+03]
            ],
        'i2': [
            [ -1.66554502e+05,   2.70933733e+05,  -1.74052246e+05,   5.43205903e+04,
               -8.75512267e+03,   5.73358984e+02,  -4.55641159e+02,  -3.95203567e-03],
            [ -2.15706075e+08,   3.46712176e+08,  -2.19342516e+08,   6.86646484e+07,
               -1.09461298e+07,   8.56547254e+05,  -1.59383063e+04,   2.24351724e+03]
            ],
        'z': [
            [  9.59579373e+03,  -1.47396976e+04,   6.94362981e+03,  -1.51139264e+03,
               -1.51535018e+02,  -1.08898155e+02,  -4.13201008e+02,  -1.79546522e-03],
            [ -3.23514338e+07,   4.70124953e+07,  -2.62797780e+07,   7.42160510e+06,
               -1.18036333e+06,   1.06569049e+05,   7.28959803e+02,   7.39449641e+03]
            ],
        'y': [
            [ -1.56137979e+05,   2.42465240e+05,  -1.47306815e+05,   4.28067607e+04,
               -6.38727141e+03,   3.73884700e+02,  -4.50571358e+02,  -2.04774887e-03],
            [ -1.88924528e+08,   2.94475198e+08,  -1.79071587e+08,   5.32635002e+07,
               -7.97354793e+06,   6.03954493e+05,  -8.92675213e+03,   2.04967838e+03]
            ],
        'N515': [
            [  1.15129313e+04,  -1.74945614e+04,   8.46294069e+03,  -1.96615600e+03,
               -8.29303133e+01,  -1.15244856e+02,  -4.16343617e+02,   1.52499844e-04],
            [ -1.97321866e+07,   2.94282093e+07,  -1.69252979e+07,   5.05172441e+06,
               -8.83913042e+05,   9.21191555e+04,   5.71543677e+02,   7.38898997e+03]
            ],
        'N656': [
            [  1.13186867e+04,  -1.71677600e+04,   8.27473180e+03,  -1.89589904e+03,
               -9.28201700e+01,  -1.14737761e+02,  -4.14296574e+02,   4.91322071e-04],
            [ -2.52048385e+07,   3.66881745e+07,  -2.05481057e+07,   5.90080624e+06,
               -9.83156724e+05,   9.71645922e+04,   3.49189805e+02,   7.60714580e+03]
            ],
        'N816': [
            [ -1.65501517e+05, 2.65225596e+05, -1.67422366e+05, 5.11417991e+04,
               -8.05597593e+03, 5.11013724e+02,  -4.54174618e+02,  -3.52735759e-03],
            [ -2.11111649e+08, 3.36261570e+08,  -2.10323885e+08,   6.49121728e+07,
               -1.01750660e+07, 7.88077371e+05,  -1.39611991e+04,   2.17954931e+03]
            ],
        'N921': [
            [ -1.66995347e+05,   2.63017929e+05,  -1.62550708e+05,   4.83455130e+04,
               -7.38661211e+03,   4.49346634e+02,  -4.52301392e+02,  -2.99761113e-03],
            [ -2.04640222e+08,   3.22255104e+08,  -1.98625679e+08,   6.01493518e+07,
               -9.20923423e+06,   7.02729746e+05,  -1.15228777e+04,   2.11290612e+03]
            ],
        'N973': [
            [  9.16226869e+03,  -1.37520999e+04,   6.05235450e+03,  -1.11685974e+03,
               -2.45192096e+02,  -9.58791049e+01,  -4.15763017e+02,  -2.73186963e-04],
            [ -3.20439555e+07,   4.87660082e+07,  -2.88299389e+07,   8.65089731e+06,
               -1.43914755e+06,   1.28246179e+05,   1.13421539e+03,   6.59765381e+03]
            ],
        }

    """
    Old parameters:

    CP_dict = {'NB0656':
               [ 6147.52560646453, -11644.9671956874, 7092.67965175815,
                 ###### -2485.58116644149, 204.389838737983, 156.280793266213,
                 -2485.58116644149, 204.389838737983, -156.280793266213,
                 -412.480162988607, -0.000473879047961772, 'NB0656'],
               'HSC-I':
               [ 17419.2342129687, -32614.9016996761, 22645.2285754691,
                 -8336.63945333745, 1374.61215520964, -271.494647513742,
                 -409.054930950492, -0.000134100887663415, 'HSC-I'],
               }

    CR_dict = {'NB0656':
               [14731353.7102142, -20958572.2574328, 11014908.2931206,
                -2240335.6606381, 7777.69740101294, 48401.0942771962,
                770.802478885535, 7737.13248666711, 'NB0656'],
               'HSC-I':
               [68873276.851416, -110287387.428125, 68613885.7163403,
                -20608617.4701426, 3013050.29930459, -187096.928825342,
                8450.36555933896, 6938.40046901406, 'HSC-I'],
               }
    """

    def getGhostsCcd(self, butler, exp, band, raFovCen, decFovCen, xFovCen_ccd, yFovCen_ccd, ra_s_d, dec_s_d, display = False, debug = False):
        """ get circle areas (x, y, r) of possible ghosts for a bright star (ra_s_d, de_s_d) in a CCD local coordinates
        """
        wcs = exp.getWcs()
        xmax_ccd = exp.getWidth() - 1.; xmin_ccd = 0.0
        ymax_ccd = exp.getHeight() -1.; ymin_ccd = 0.0

        cdMatrix = wcs.getCDMatrix() # if solvetansip is run, wcs should be same as wcs49
        cd1_1 = cdMatrix[0][0]
        cd1_2 = cdMatrix[0][1]
        cd2_1 = cdMatrix[1][0]
        cd2_2 = cdMatrix[1][1]

#        pscale = numpy.sqrt(numpy.absolute(cd1_1*cd2_2-cd1_2*cd2_1))
        pscale = wcs.pixelScale().asDegrees()

        theta, phi = self.sphr_trig(raFovCen.asDegrees(), decFovCen.asDegrees(), ra_s_d, dec_s_d) # this can be replaced with angularSeparation later.  crval should come from CCD49 center coords(+applied offsets above)
        phi += 90.0
        if phi >= 360.0:
            phi = phi - 360.0

        if debug:
            print 'ghost star (ra,dec,theta,phi; w.r.t. fovcen):', ra_s_d, dec_s_d, theta, phi

        CP = self.CP_CR_dict[band][0] # coeff for position
        CR = self.CP_CR_dict[band][1] # coeff for size(radius)

        # Ghost map
        ghostCcdList = []
        if theta <= 0.5: # applicable within a area of r=0.5 deg only

            midpos = CP[0]*theta**7 + CP[1]*theta**6 + CP[2]*theta**5 + CP[3]*theta**4 + CP[4]*theta**3 + CP[5]*theta**2 + CP[6]*theta + CP[7]   # 7th order fit

            xm = midpos/self._pixscaleMm*numpy.cos(numpy.radians(phi))
            ym = midpos/self._pixscaleMm*numpy.sin(numpy.radians(phi))
            deta, dxi = self.delta_radec(midpos/self._pixscaleMm*pscale, phi+90.0, decFovCen.asDegrees())
            dpx0 = deta/pscale
            dpy0 = dxi/pscale
            if debug:
                print 'ghost (xm,ym; rel.axis)', xm, ym
                print 'ghost (deta,dxi; rel.axis)', deta, dxi
                print 'ghost (dpx0,dpy0; rel.axis)', dpx0, dpy0
            dpx = (-cd1_1*xm+cd1_2*ym)/pscale
            dpy = (-cd2_1*xm+cd2_2*ym)/pscale

            rad = CR[0]*theta**7 + CR[1]*theta**6 + CR[2]*theta**5 + CR[3]*theta**4 + CR[4]*theta**3 + CR[5]*theta**2 + CR[6]*theta + CR[7]   # 7th order fit
            rad /= (self._pixscaleMm * 1000.0)

            if debug:
                print 'ghost (midpos,dr,dd,rad; rel.axis)', midpos, dpx, dpy, rad
                print 'ghost (x,y,rad; rel.orgfits.cood)', dpx+xFovCen_ccd, dpy+yFovCen_ccd, rad

            x_local, y_local = dpx + xFovCen_ccd, dpy + yFovCen_ccd

            # ensure this ghost prediction circle overlaps with at least a part of the CCD pixels
            if debug:
                print "(%f, %f) - (%f, %f, %f, %f)" % (xmax_ccd, ymax_ccd, x_local + rad, x_local - rad, y_local + rad, y_local + rad)

            if ((x_local + rad) > xmin_ccd) and ((x_local - rad) < xmax_ccd) and \
                    ((y_local + rad) > ymin_ccd) and ((y_local - rad) < ymax_ccd):

                ghostCcdList.append([x_local, y_local, rad])

        return ghostCcdList

    def run(self, butler, band, wcs, raFovCen, decFovCen, xFovCen_ccd, yFovCen_ccd, ra_s_d, de_s_d, display = False, outputReg = True):
         self.getGhostsCcd(self, butler, exp, band, wcs, raFovCen, decFovCen, xFovCen_ccd, yFovCen_ccd, ra_s_d, de_s_d, display = display, outputReg = outputReg)


class HscMustacheGhostMaskConfig(HscMustacheGhostTask.ConfigClass):
    astrometry = pexConfig.ConfigField(
        dtype = Astrometry.ConfigClass,
        doc = "Configuration for the astrometry",
    )

class HscMustacheGhostMaskTask(HscMustacheGhostTask):

    ConfigClass = HscMustacheGhostMaskConfig
    _DefaultName = 'mask_mustache'

    def run(self, dataRef, exp, centerCoord = None, outputReg = False, display = False):
        #exp = afwImage.ExposureF(fileName)
        #astrom = measAst.Astrometry(config = MeasAstromConfig())
        #butler = dafPersist.ButlerFactory(mapper=obsHsc.HscMapper(root = root)).create()
        self.astrom = Astrometry(self.config.astrometry)
        butler = dataRef.getButler()
        camera = butler.get("camera")
        raft = camera[0]
        raft = cameraGeom.cast_Raft(raft)

        md = exp.getMetadata()
        #ccdId = int(md.get('DET-ID'))
        ccdId = dataRef.dataId['ccd']

        # getting coords of FOV center
        if not centerCoord:
            raFovCenTel = md.get('RA') # FOV center telescope commanded
            decFovCenTel = md.get('DEC')
            raFovCenDeg, decFovCenDeg = sex2dec(raFovCenTel, decFovCenTel)
            raFovCen = numpy.radians(raFovCenDeg) * afwGeom.radians
            decFovCen = numpy.radians(decFovCenDeg) * afwGeom.radians

            centerCoord = afwCoord.Coord(afwGeom.Point2D(raFovCen.asDegrees(), decFovCen.asDegrees()))

        raFovCen, decFovCen = centerCoord

        rad = numpy.radians(self.config.maxRadius) * afwGeom.radians
        filterName = exp.getFilter().getName()

        cat = self.astrom.getReferenceSources(raFovCen, decFovCen, rad, self.config.filterNameRef[filterName])

        schema = cat.getSchema()
        key_flux = schema.find("%s" % self.config.filterNameRef[filterName]).key
        key_coord =  schema.find("coord").key
        catBright = [] 
        for s in cat:
            coord = s.get(key_coord)
            dist = coord.angularSeparation(centerCoord).asDegrees()
            if self.config.minRadius < dist and dist < self.config.maxRadius:
                flux = s.get(key_flux)
                mag = -2.5*numpy.log10(flux)
                if -2.0 < mag and mag < self.config.maglim[filterName] :
                    catBright.append(s)
        del cat

        # getting (x, y) list of possible 'mustache' ghosts
        mjd = float(md.get('MJD'))
        ra_t_d = raFovCen.asDegrees()
        de_t_d = decFovCen.asDegrees()
        inr_d = float(md.get('INR-STR'))

        ghostCcdList = []
        for s in catBright:
            ra, dec = s.get(key_coord)
            ra_s_d = ra.asDegrees()
            de_s_d = dec.asDegrees()
            ghosts = self.getGhostsCcd(butler, exp, self.config.filterNameRef[filterName], ccdId, mjd, ra_t_d, de_t_d, inr_d, ra_s_d, de_s_d, outputReg = outputReg, display = display)
            ghostCcdList += ghosts


        if display:
            import lsst.afw.display.ds9 as ds9 
            # ds9.mtv(exp.getMaskedImage().getImage(), frame = frame, title = fileName)
            ds9.mtv(exp.getMaskedImage().getImage())
            for x_local, y_local in ghostCcdList:
                # ds9.dot('x', x_local, y_local, size=20.0, ctype='magenta', frame = frame)
                ds9.dot('x', x_local, y_local, size=20.0, ctype='magenta')

        if outputReg:
            regFileName = 'ghosts_mustache.reg'
            with open(regFileName, 'w') as f:
                f.write('global color=magenta\n')
                f.write('fk5\n')
                for x_local, y_local in ghostCcdList:
                    f.write('image;point(%s,%s) # point=x 20\n' % (x_local, y_local))
                    #print 'image;point(%s,%s) # point=x 20' % (x_local, y_local)

        mask = exp.getMaskedImage().getMask()
        goodMask = mask.getPlaneBitMask("DETECTED")
        badMask = mask.getPlaneBitMask(["NO_DATA", "EDGE", "SAT", "SUSPECT", "INTRP", "BAD"])
        self.maskGhostImage(mask, ghostCcdList, goodMask, badMask)


    def maskGhostImage(self, mask, ghostCcdList, goodMask, badMask):
        """ mask ghosts in image """

        # mask possible ghosts' footprints with 'GHOST' bit
        mask.addMaskPlane('GHOST_MUSTACHE')
        ghostBit = mask.getPlaneBitMask('GHOST_MUSTACHE') | mask.getPlaneBitMask('BAD') # for test
        detectedBit = mask.getPlaneBitMask('DETECTED')

        fpList = self.maskGhostFootprints(mask, ghostCcdList, detectedBit, ghostBit)

        return fpList

        #exp.writeFits('masked.fits')

    def maskGhostFootprints(self, mask, xyList, detectedBit, ghostBit):
        """ select footprints of possible ghosts and set new mask bit to them 
        """

        thresh = afwDetection.Threshold(detectedBit, afwDetection.Threshold.BITMASK)
        fpList = afwDetection.FootprintSet(mask, thresh).getFootprints()

        xmin, ymin = mask.getBBox(afwImage.PARENT).getMin()

        fpGhostList = self.getGhostFootprints(mask, fpList, xyList)

        # drop duplicated footprints
        print 'fpGhostList raw: %s' % len(fpGhostList)
        fpGhostList = list(set(fpGhostList))
        print 'fpGhostList duplication dropped: %s' % len(fpGhostList)

        for fp in fpGhostList:
            for s in fp.getSpans():
                x0, x1 = s.getX0() - xmin, s.getX1() - xmin
                y = s.getY() - ymin
    #            x0, x1 = s.getX0(), s.getX1()
    #            y = s.getY()
                for x in range(x0, x1):
                    mask.set(x, y, mask.get(x, y) | ghostBit)

        return fpGhostList

    def getGhostFootprints(self, mask, fpList, xyList):
        """ select footprints of ghost candidates from the input footprint list
        """

        # check if this footprint overlaps with the ghosts candidates
        xmin, ymin = mask.getBBox(afwImage.PARENT).getMin()

        fpGhostList = []
        for fp in fpList:

            # footprint should be larger than the given minArea to be a ghost
            if fp.getArea() < self.config.minGhostArea:
                continue

            # check if this fp overlaps with ghost candidates
            # define ghost regions with margin for recognizing this footprint as a candidate
            for xGhost, yGhost in xyList:
                for i in (-3., -2., -1., 0., 1., 2., 3.):
                    xGhost_tmp = xGhost + i * self.config.marginGhostPosition
                    if fp.contains(afwGeom.Point2I(int(xGhost_tmp), int(yGhost))):
                        #print xGhost_tmp, yGhost
                        fpGhostList.append(fp)
                        break

        return fpGhostList


class HscCausticGhostMaskConfig(HscCausticGhostTask.ConfigClass):

    astrometry = pexConfig.ConfigField(
        dtype = Astrometry.ConfigClass,
        doc = "Configuration for the astrometry",
    )
    useExternalCatalog = pexConfig.Field(dtype = bool, default = True,
                                doc = "use external star catalog for bright sources. if False, detected sources are used.")


class HscCausticGhostMaskTask(HscCausticGhostTask):

    ConfigClass = HscCausticGhostMaskConfig
    _DefaultName = 'mask_caustic'

    def run(self, dataRef, exp, centerCoord = None, outputReg = False, display = False):

        butler = dataRef.getButler()
        camera = butler.get("camera")
        raft = camera[0]
        raft = cameraGeom.cast_Raft(raft)

        md = exp.getMetadata()
        wcs = exp.getWcs()

        #ccdId = int(md.get('DET-ID'))
        ccdId = dataRef.dataId['ccd']

        # getting coords of FOV center
        if not centerCoord:
            raFovCenTel = md.get('RA') # FOV center telescope commanded
            decFovCenTel = md.get('DEC')
            raFovCenDeg, decFovCenDeg = sex2dec(raFovCenTel, decFovCenTel)
            raFovCen = numpy.radians(raFovCenDeg) * afwGeom.radians
            decFovCen = numpy.radians(decFovCenDeg) * afwGeom.radians

            centerCoord = afwCoord.Coord(afwGeom.Point2D(raFovCen.asDegrees(), decFovCen.asDegrees()))

        raFovCen, decFovCen = centerCoord

        xFovCen_ccd, yFovCen_ccd = wcs.skyToPixel(centerCoord) # fov center in this ccd pixel coords

        rad = numpy.radians(self.config.maxRadius) * afwGeom.radians
        filterName = exp.getFilter().getName()

        if self.config.useExternalCatalog:
            astrom = Astrometry(self.config.astrometry)
            cat = astrom.getReferenceSources(raFovCen, decFovCen, rad, self.config.filterNameRef[filterName])
            schema = cat.getSchema()
            key_flux = schema.find("%s" % self.config.filterNameRef[filterName]).key
            key_coord =  schema.find("coord").key
            magzero = 0.0
        else: # use src
            raise NotImplementedError()
            # here, we should collect bright stars from all field
            schema = cat.getSchema()
            key_flux = schema.find('flux.psf').key
            key_coord =  schema.find("coord").key
            magzero = 2.5 * numpy.log10(exp.getCalib().getFluxMag0()[0])

        catBright = [] 
        for s in cat:
            coord = s.get(key_coord)
            dist = coord.angularSeparation(centerCoord).asDegrees()
            if self.config.minRadius <= dist and dist < self.config.maxRadius:
                flux = s.get(key_flux)
                mag = -2.5*numpy.log10(flux) + magzero
                # print s.getId(), mag, coord[0].asDegrees(), coord[1].asDegrees(), dist
                if self.config.magBright[filterName] < mag and mag < self.config.magFaint[filterName] :
                    catBright.append(s)
        del cat

        ghostCcdList = []
        for s in catBright:
            ra, dec = s.get(key_coord)
            ra_s_d = ra.asDegrees()
            dec_s_d = dec.asDegrees()

            ghosts = self.getGhostsCcd(butler, exp, filterName, raFovCen, decFovCen, xFovCen_ccd, yFovCen_ccd, ra_s_d, dec_s_d, display = display)
            ghostCcdList += ghosts

        if display:
            import lsst.afw.display.ds9 as ds9 
            # ds9.mtv(exp.getMaskedImage().getImage(), frame = frame, title = fileName)
            ds9.mtv(exp.getMaskedImage().getImage())
            for x_local, y_local, rad in ghostCcdList:
                # ds9.dot('x', x_local, y_local, size=20.0, ctype='magenta', frame = frame)
                ds9.dot('o', x_local, y_local, size=rad, ctype='magenta')

        if outputReg:
            regFileName = 'ghosts_caustic.reg'
            with open(regFileName, 'w') as f:
                f.write('global color=magenta\n')
                f.write('fk5\n')
                for x_local, y_local, rad in ghostCcdList:
                    f.write('image;circle(%s,%s,%.1f)\n' % (x_local, y_local, rad))
                    print 'image;circle(%s,%s,%.1f)' % (x_local, y_local, rad)

        mask = exp.getMaskedImage().getMask()
        goodMask = mask.getPlaneBitMask("DETECTED")
        badMask = mask.getPlaneBitMask(["NO_DATA", "EDGE", "SAT", "SUSPECT", "INTRP", "BAD"])

        self.maskGhostImage(mask, ghostCcdList, goodMask, badMask)


    def maskGhostImage(self, mask, ghostCcdList, goodMask, badMask):
        """ mask ghosts in image """

        # mask possible ghosts' footprints with 'GHOST' bit
        mask.addMaskPlane('GHOST_CAUSTIC')
        ghostBit = mask.getPlaneBitMask('GHOST_CAUSTIC') | mask.getPlaneBitMask('BAD') # for test
        detectedBit = mask.getPlaneBitMask('DETECTED')

        fpList = self.maskGhostFootprints(mask, ghostCcdList, detectedBit, ghostBit)

        #exp.writeFits('masked.fits')

    def maskGhostFootprints(self, mask, xyrList, detectedBit, ghostBit, debug = True):
        """ select footprints of possible ghosts and set new mask bit to them 
        """

        thresh = afwDetection.Threshold(detectedBit, afwDetection.Threshold.BITMASK)
        fpList = afwDetection.FootprintSet(mask, thresh).getFootprints()

        xmin, ymin = mask.getBBox(afwImage.PARENT).getMin()

        fpGhostList = self.getGhostFootprints(mask, fpList, xyrList)

        # drop duplicated footprints
        print 'fpGhostList raw: %s' % len(fpGhostList)
        fpGhostList = list(set(fpGhostList))
        print 'fpGhostList after duplication drop: %s' % len(fpGhostList)

        for fp in fpGhostList:
            if debug:
                print 'fpGhost area: %s' % fp.getArea()
            for s in fp.getSpans():
                x0, x1 = s.getX0() - xmin, s.getX1() - xmin
                y = s.getY() - ymin
                for x in range(x0, x1):
                    mask.set(x, y, mask.get(x, y) | ghostBit)

        return fpGhostList

    def getGhostFootprints(self, mask, fpList, xyrList):
        """ select footprints of ghost candidates from the input footprint list
        """

        # check if this footprint overlaps with the ghosts candidates
        xmin, ymin = mask.getBBox(afwImage.PARENT).getMin()

        fpGhostList = []
        for fp in fpList:

            # footprint should be larger than the given minArea to be a ghost
            if fp.getArea() < self.config.minGhostArea:
                continue

            # check if this fp overlaps with a ghost prediction circle
            # define ghost regions with margin for recognizing this footprint as a ghost candidate

            for xGhost, yGhost, radius in xyrList:

                # generate a set of grids uniformly placed within this ghost candidate circle
                grid_size = 10.0 # pix
                for yc in numpy.arange(yGhost-radius, yGhost+radius, grid_size):
                    xmin = xGhost - numpy.sqrt(radius**2. - (yc - yGhost)**2.)
                    xmax = xGhost + numpy.sqrt(radius**2. - (yc - yGhost)**2.)
                    for xc in numpy.arange(xmin, xmax, grid_size):

                        if fp.contains(afwGeom.Point2I(int(xc), int(yc))):
                            #print xGhost_tmp, yGhost
                            fpGhostList.append(fp)
                            break

        return fpGhostList


    def sphr_trig(self, a0, d0, a1, d1):
        a_cos = numpy.cos(numpy.radians(90.0-d1))*numpy.cos(numpy.radians(90.0-d0)) + numpy.sin(numpy.radians(90.0-d1))*numpy.sin(numpy.radians(90.0-d0))*numpy.cos(numpy.radians(a1-a0))
        theta = numpy.degrees(numpy.arccos(numpy.absolute(a_cos)))
        if a_cos < 0.0:
            theta = 180.0 - theta
        c_sin = numpy.sin(numpy.radians(90.0-d1))*numpy.sin(numpy.radians(a1-a0))/numpy.sin(numpy.radians(theta))
        c_cos = (numpy.cos(numpy.radians(90.0-d1))*numpy.sin(numpy.radians(90.0-d0)) - numpy.sin(numpy.radians(90.0-d1))*numpy.cos(numpy.radians(90.0-d0))*numpy.cos(numpy.radians(a1-a0)))/numpy.sin(numpy.radians(theta))
        phi = numpy.degrees(numpy.arctan2(c_sin, c_cos))

        return theta, phi

    def delta_radec(self, theta, phi, d0):
        z_cos = numpy.cos(numpy.radians(theta))*numpy.cos(numpy.radians(90.0-d0)) + numpy.sin(numpy.radians(theta))*numpy.sin(numpy.radians(90.0-d0))*numpy.cos(numpy.radians(phi))
        z = numpy.degrees(numpy.arccos(z_cos))
        #if z_cos < 0.0:
        #    z = 180.0 - z
        da_sin = numpy.sin(numpy.radians(theta))*numpy.sin(numpy.radians(phi))/numpy.sin(numpy.radians(z))
        da = numpy.degrees(numpy.arcsin(da_sin))

        #print da, d0, z
        return da, (z+d0-90.0)


def sex2dec(ra_str, dec_str):
    """ convert string sexagesimal (ra, dec) to (ra, dec) in degrees 
    """

    ra_str = ra_str.strip()
    dec_str = dec_str.strip()

    if ra_str[0] == '-':
        raise 'invalid ra: %s' % ra_str

    ra = map(float, ra_str.split(':'))
    outra = (ra[0] + ra[1] / 60. + ra[2] / 3600.) * 15.0
    dec = map(float, dec_str.split(':'))

    if dec_str[0] == '-':
        outdec = dec[0] - dec[1] / 60. - dec[2] / 3600.
    else:
        outdec = dec[0] + dec[1] / 60. + dec[2] / 3600.
    return outra, outdec



